/*
 * certkey_engine.c
 *
 * - Holds the long-term certified keys.
 * - Performs the signature and verification.
 */

#include "seal.h"

static EVP_PKEY* CERTKEY_pubkey_retrieve(CERTKEY_control_p_t certkey_engine, 
		char const* uid);
static int CERTKEY_guid_to_uid(char const* guid, char* uid);

/*
 * Initialize the certkey structure
 */
CERTKEY_control_p_t CERTKEY_init(char const* my_uid, char* my_passwd) {
	
	CERTKEY_control_p_t certkey_engine;

	FILE* fp;
	char file_name[SEAL_UID_LEN + 20];
	X509* cert;

	certkey_engine = (CERTKEY_control_p_t)malloc(sizeof(CERTKEY_control_t));

	// init the pubkey structure
	certkey_engine->pubkey = (CERTKEY_pubkey_p_t)malloc(sizeof(
				CERTKEY_pubkey_t));
	certkey_engine->pubkey->head = NULL;
	certkey_engine->pubkey->tail = NULL;
   
	// dynamic loading of algorithms
	EVP_add_cipher(EVP_des_ede3_cbc());

	// obtain long-term private key	
	sprintf(file_name, "%s/%s_key.pem", CERTKEY_KEY_DIR, my_uid);
	if ((fp = fopen(file_name, "r")) == NULL) {
		debug_err("cannot open private key\n");
		exit(-1);
	}
	if ((certkey_engine->cert_pri_key = PEM_read_PrivateKey(fp, NULL, NULL, 
					my_passwd)) == NULL) {
		debug_err("cannot read private key. Check your passphrase\n");
		exit(-1);
	}
	fclose(fp);

	// obtain certified public key
	sprintf(file_name, "%s/%s_cert.pem", CERTKEY_CERT_DIR, my_uid);
	if ((fp = fopen(file_name, "r")) == NULL) {
		debug_err("cannot open x509 cert\n");
		exit(-1);
	}
	if ((cert = PEM_read_X509(fp, NULL, NULL, NULL)) == NULL) {
		debug_err("cannot read my cert file\n");
		exit(-1);
	}
	if ((certkey_engine->cert_pub_key = X509_get_pubkey(cert)) == NULL) {
		debug_err("cannot read x509's public key\n");
		exit(-1);
	}
	fclose(fp);	
	
	// init the lock
	pthread_mutex_init(&(certkey_engine->lock), NULL);

	// release the loading
	EVP_cleanup();

	return certkey_engine;
}

/*---------------------------------------------------------------------------
 *					PUBKEY MANAGEMENT
 *---------------------------------------------------------------------------*/

/*
 * Install all entries to pubkey list.
 */
void CERTKEY_pubkey_install(CERTKEY_control_p_t certkey_engine, 
		char* const* uids, int no_member) {
	
	CERTKEY_pubkey_entry_p_t new_entry_p;	// new pubkey entry
	FILE* fp;								// cert's file pointer
	char file_name[50];						// cert's file name
	X509* cert;								// certificate
	int i;									// loop counter


	for (i=0; i<no_member; ++i) {
		// malloc new entry
		if ((new_entry_p = (CERTKEY_pubkey_entry_p_t)malloc(
						sizeof(CERTKEY_pubkey_entry_t))) == NULL) {
			debug_err("Unable to create certkey_pubkey_entry at i=%d\n", i);
			return;
		}

		// set the uid
		memset(new_entry_p, 0, sizeof(CERTKEY_pubkey_entry_t));
		memcpy(new_entry_p->uid, uids[i], SEAL_UID_LEN);
		
		// extract the certificate and public key
		sprintf(file_name, "%s/%s_cert.pem", CERTKEY_CERT_DIR, uids[i]);
		if ((fp = fopen(file_name, "r")) == NULL) {
			debug_err("cannot open x509 cert file: %s\n", file_name);
			exit(-1);
		}
		if ((cert = PEM_read_X509(fp, NULL, NULL, NULL)) == NULL) {
			debug_err("cannot read cert file: %s\n", file_name);
			exit(-1);
		}
		if ((new_entry_p->pub_key = X509_get_pubkey(cert)) == NULL) {
			debug_err("cannot read x509's public key: %s\n", file_name);
			exit(-1);
		}
		fclose(fp);	

		// insert the new entry at the list end
		new_entry_p->prev = certkey_engine->pubkey->tail;
		new_entry_p->next = NULL;
		if (certkey_engine->pubkey->head == NULL) {
			certkey_engine->pubkey->head = new_entry_p;
		} else {
			certkey_engine->pubkey->tail->next = new_entry_p;
		}
		certkey_engine->pubkey->tail = new_entry_p;
	}
}

/*
 * Check if the pubkey exists. Used to check if a member is valid
 */
int CERTKEY_pubkey_is_exist(CERTKEY_control_p_t certkey_engine, 
		char const* guid) {

	CERTKEY_pubkey_entry_p_t tmp_entry_p;
	char uid[SEAL_UID_LEN];

	// convert the sender to sender uid
	if (CERTKEY_guid_to_uid(guid, uid) == 0) {	
		debug_err("Incorrect GUID in pubkey_is_exist\n");
		exit(-1);
	}

	tmp_entry_p = certkey_engine->pubkey->head;
	while (tmp_entry_p != NULL) {
		if (memcmp(tmp_entry_p->uid, uid, SEAL_UID_LEN) == 0) {
			return 1;
		}
		tmp_entry_p = tmp_entry_p->next;
	}

	// reaching 0 (pubkey entry not found)
	return 0;
}

/*---------------------------------------------------------------------
 *					SIGNATURE / VERIFICATION 
 *---------------------------------------------------------------------*/

/*
 * Sign the given message. Return 1 on success, 0 on error.
 */
int CERTKEY_sign(CERTKEY_control_p_t certkey_engine, void const* msg,
		int msg_len, char* sig) {

	EVP_MD_CTX evp_md_ctx;				// message digest context
	int sig_len;
	
	// dynamic loading of digest algorithm	
	EVP_add_digest(EVP_sha1());
	
	EVP_SignInit(&evp_md_ctx, EVP_sha1());
	EVP_SignUpdate(&evp_md_ctx, msg, msg_len);
	if (EVP_SignFinal(&evp_md_ctx, sig, &sig_len, 
				certkey_engine->cert_pri_key) == 0) {
		// release the loading
		EVP_cleanup();
		return 0;					// signature failed
	}
	if (sig_len != CERTKEY_LEN) {		
		// release the loading
		EVP_cleanup();
		// debugging, normally sig_len == CERTKEY_LEN
		debug("Sig_len: %d\n", sig_len);
		return 0;
	}
	
	// release the loading
	EVP_cleanup();

	return 1;	
}

/*
 * Verify the given signature. Return 1 on success, 0 on error.
 */
int CERTKEY_verify(CERTKEY_control_p_t certkey_engine, char const* sender,
		void const* msg, int msg_len, char* sig, int sig_len) {

	EVP_MD_CTX evp_md_ctx;			// message digest context
	EVP_PKEY* sender_pub_key;		// sender's certified public key

	char sender_uid[SEAL_UID_LEN];

	// dynamic loading of digest algorithm	
	EVP_add_digest(EVP_sha1());
	
	// convert the sender to sender uid
	if (CERTKEY_guid_to_uid(sender, sender_uid) == 0) {	
		debug_err("Incorrect GUID in CERTKEY_verify\n");
		exit(-1);
	}

	// retrieve the public key
	sender_pub_key = CERTKEY_pubkey_retrieve(certkey_engine, sender_uid);
	if (sender_pub_key == NULL) {
		// release the loading
		EVP_cleanup();
		return 0;
	}

	EVP_VerifyInit(&evp_md_ctx, EVP_sha1());
	EVP_VerifyUpdate(&evp_md_ctx, msg, msg_len);
	if (EVP_VerifyFinal(&evp_md_ctx, sig, sig_len, sender_pub_key) == 0) {
		// release the loading
		EVP_cleanup();
		// verification failed	
		return 0;
	}
	// release the loading
	EVP_cleanup();
	return 1;
}

/*---------------------------------------------------------------------------
 *					DESTROY
 *---------------------------------------------------------------------------*/

/*
 * Destroy the certkey structure
 */
void CERTKEY_destroy(CERTKEY_control_p_t certkey_engine) {

	CERTKEY_pubkey_entry_p_t del_node_p;

	// delete all pub_keys in the pubkey list
	while (certkey_engine->pubkey->head != NULL) {
		del_node_p = certkey_engine->pubkey->head;
		certkey_engine->pubkey->head = del_node_p->next;
		EVP_PKEY_free(del_node_p->pub_key);
		free(del_node_p);
	}
	free(certkey_engine->pubkey);

	// free my cert_pri_key and cert_pub_key
	EVP_PKEY_free(certkey_engine->cert_pri_key);
	EVP_PKEY_free(certkey_engine->cert_pub_key);

	// free the certkey engine	
	free(certkey_engine);
}

/*---------------------------------------------------------------------
 *                     STATIC FUNCTIONS
 *---------------------------------------------------------------------*/

/*
 * Retrieve the corresponding public key for the given uid
 */
static EVP_PKEY* CERTKEY_pubkey_retrieve(CERTKEY_control_p_t certkey_engine, 
		char const* uid) {
	
	CERTKEY_pubkey_entry_p_t tmp_entry_p;

	tmp_entry_p = certkey_engine->pubkey->head;
	while (tmp_entry_p != NULL) {
		if (memcmp(tmp_entry_p->uid, uid, SEAL_UID_LEN) == 0) {
			return tmp_entry_p->pub_key;
		}
		tmp_entry_p = tmp_entry_p->next;
	}

	// reaching null (pubkey entry not found)
	return NULL;
}


/*
 * Convert guid to uid
 * Return 1 on success, 0 on failure
 */
static int CERTKEY_guid_to_uid(const char* guid, char* uid) {
	char* ptr_pound;

	// initialize uid
	memset(uid, 0, SEAL_UID_LEN);

	// remove the 1st '#' sign, and the trailing 2-digit timestamp values
	ptr_pound = strpbrk(guid + 1, "#");
	if (ptr_pound != NULL && ptr_pound - guid  > 3 &&
			ptr_pound - guid - 3 < SEAL_UID_LEN) {
		memcpy(uid, guid + 1, ptr_pound - guid - 3);
	} else {
		debug_err("Incorrect GUID\n");
		return 0;
	}
	return 1;
}
