/*
 * gauger.c
 *
 * - gauger program
 *		This program runs automatically and joins / leaves the system
 *		following an exponential distributed time plus a constant. 
 *		The constant is to allow some buffer period between two events.
 */

#include "gauger.h"

// log file pointer
static FILE* log_fp_g;

// lock for all threads (main_thread, exit_thread and stream_thread)
static pthread_mutex_t thread_lock_g = PTHREAD_MUTEX_INITIALIZER;

// check inside/outside the group;
static int in_out_flag_g = LEFT;

// sleep time genenerated randomly / by pattern file
static int gen_flag_g = PTN_RANDOM;

// function prototypes
static void* exit_handler();
static void* stream_handler();
static void* msg_recv_handler();
static void print_usage(char*);

/*
 * main program
 */
int main(int argc, char** argv) {
	pthread_cond_t dummy_cond;              // dummy condition variable
	struct timeval tv;                      // time structures
	struct timespec ts;
	double avg_in_time, avg_out_time;
	double sleep_time = -1.0;

	pthread_t exit_thread = 0;				// * a thread determining the end
	pthread_t stream_thread = 0;			// * a thread streaming out data
	pthread_t msg_recv_thread = 0;			// * a thread receiving msgs (only 
											//   when some gaugers stream msgs)

	FILE* ptn_fp;							// pattern file pointer
	double ptn_in_time[PTN_MAX_ENTRIES];
	double ptn_out_time[PTN_MAX_ENTRIES];
	int ptn_in_cnt = 0;
	int ptn_out_cnt = 0;

	char readline[50];						// line read from the file

	char password[10];

	int i;									// loop counter

	// start-up	
	printf("Start....\n");

	// init the seal session object
	seal_session_g = SEAL_init();

	// set password	of the seal session object
	memset(password, 0, sizeof(password));	
	if (strlen(seal_session_g->uid) == 3) {
		sprintf(password, "%s%d", seal_session_g->uid, 1);
	} else {
		sprintf(password, "%s", seal_session_g->uid);
	}
	SEAL_set_passwd(seal_session_g, password);

	// initialize the gauger engine
	gauger_g = GAUGER_init();

	// initialize the log file pointer
	log_fp_g = stdout;

	// parse options from arugments 
	if (argc == 1) {
		// set the default values
		avg_in_time = 60.0;
		avg_out_time = 60.0;
		log_fp_g = stdout;
		gen_flag_g = PTN_RANDOM;

	} else {		// argc > 1 
		
		if (strncmp(argv[1], "-p", 2) == 0) {
			if (argc == 4) {
				// open the pattern file
				if ((ptn_fp = fopen(argv[2], "r")) == NULL) {
					debug_err("Cannot open pattern file: %s\n", argv[2]);
					exit(-1);
				}
				// read from the pattern file
				memset(ptn_in_time, 0, PTN_MAX_ENTRIES * sizeof(double));
				memset(ptn_out_time, 0, PTN_MAX_ENTRIES * sizeof(double));
				memset(readline, 0, sizeof(readline));
				// 1) get the avg_in_time and avg_out_time
				fgets(readline, sizeof(readline), ptn_fp);	
				sscanf(readline, "%lf\t%lf\n", &avg_in_time, &avg_out_time);	
				// 2) start reading the pattern	
				memset(readline, 0, sizeof(readline));
				i = 0;
				while (fgets(readline, sizeof(readline), ptn_fp) != NULL
						&& i < PTN_MAX_ENTRIES) {
					sscanf(readline, "%lf\t%lf\n", &ptn_in_time[i], 
							&ptn_out_time[i]);
					memset(readline, 0, sizeof(readline));
					i++;
				}
				// 3) close the pattern file
				fclose(ptn_fp);

				// initialize the output file	
				log_fp_g = fopen(argv[3], "w");
				
				// initialize gen_flag_g	
				gen_flag_g = PTN_PREGEN;

			} else {
				print_usage(argv[0]);
				exit(-1);
			}
	
		} else if (strncmp(argv[1], "-r", 2) == 0) {
			if (argc == 4) {
				// set the average times and default values
				avg_in_time = atof(argv[2]);
				avg_out_time = atof(argv[3]);
				log_fp_g = stdout;
				gen_flag_g = PTN_RANDOM;
			
			} else {
				print_usage(argv[0]);
				exit(-1);
			}

		} else if (strncmp(argv[1], "-s", 2) == 0) {
			if (argc == 3) {
				// create thread and set the average times and default values
				pthread_create(&stream_thread, NULL, (void*)stream_handler,
						(void*)atoi(argv[2]));
				avg_in_time = 60.0;
				avg_out_time = 60.0;
				log_fp_g = stdout;
				gen_flag_g = PTN_RANDOM;

			} else {
				print_usage(argv[0]);
				exit(-1);
			}

		} else if (strncmp(argv[1], "-h", 2) == 0) {
			print_usage(argv[0]);
			exit(0);

		} else {		// argc > 1 but no option matched
			print_usage(argv[0]);
			exit(-1);
		}

	}		// end: argc > 1
	
	
	// create the exit thread
	pthread_create(&exit_thread, NULL, (void*)&exit_handler, NULL);

	// set the timer for join/leave
	pthread_cond_init(&dummy_cond, NULL);
	gettimeofday(&tv, NULL);                // get the current time
	ts.tv_sec = tv.tv_sec;
	ts.tv_nsec = tv.tv_usec * 1000;

	// set the rand seed
	srand(time(NULL));

	// Determine how long the gauger stays outside the group	
	if (gen_flag_g == PTN_RANDOM) {
		sleep_time = LEAST_SLEEP_TIME + (- avg_out_time * 
				log(1.0 - (double)rand() / RAND_MAX));
	} else if (gen_flag_g == PTN_PREGEN) {
		sleep_time = LEAST_SLEEP_TIME + ptn_out_time[ptn_out_cnt++];
	}
	
	debug("Prepare to join the group after time %f\n", sleep_time);	
	ts.tv_sec = ts.tv_sec + (int)sleep_time;
	ts.tv_nsec = (int)(sleep_time / 1000000000);
	if (ts.tv_nsec > 1000000000) {
		ts.tv_sec++;
		ts.tv_nsec -= 1000000000;
	}
	pthread_cond_timedwait(&dummy_cond, &thread_lock_g, &ts);		
	
	in_out_flag_g = JOINED;	


	// join/leave according to the arrival/leave rate	
	while (1) {
		if (in_out_flag_g == JOINED) {
			
			debug("Join the group at time %d\n", time(NULL));
			
			// Join the group
			if (!SEAL_join(seal_session_g, "mygroup")) {
				exit(-1);
			}
		
			// start the msg recv thread, wait for any incoming messages
			pthread_create(&msg_recv_thread, NULL, (void*)&msg_recv_handler,
					NULL);

			// Determine how long the gauger stays in the group	
			if (gen_flag_g == PTN_RANDOM) {
				sleep_time = LEAST_SLEEP_TIME + (- avg_in_time * 
						log(1.0-(double)rand()/RAND_MAX));
			} else if (gen_flag_g == PTN_PREGEN) {
				sleep_time = LEAST_SLEEP_TIME + 
					ptn_in_time[ptn_in_cnt++];
			}
			
			debug("Join the group for time %f\n", sleep_time);	
			
			ts.tv_sec = ts.tv_sec + (int)sleep_time;
			ts.tv_nsec = (int)(sleep_time / 1000000000);
			if (ts.tv_nsec > 1000000000) {
				ts.tv_sec++;
				ts.tv_nsec -= 1000000000;
			}
			pthread_cond_timedwait(&dummy_cond, &thread_lock_g, &ts);	

			in_out_flag_g = LEFT;

		} else {

			debug("Leave the group at time %d\n", time(NULL));
	
			// cancel and join the msg recv thread
			pthread_cancel(msg_recv_thread);
			pthread_join(msg_recv_thread, NULL);

			// Leave the group
			if (!SEAL_leave(seal_session_g)) {
				exit(-1);
			}

			// reset the gauger engine
			GAUGER_reset(gauger_g);
			
			// Determine how long the gauger stays outside the group
			if (gen_flag_g == PTN_RANDOM) {
				sleep_time = LEAST_SLEEP_TIME + (- avg_out_time * 
						log(1.0 - (double)rand() / RAND_MAX));
			} else if (gen_flag_g == PTN_PREGEN) {
				sleep_time = LEAST_SLEEP_TIME + 
					ptn_out_time[ptn_out_cnt++];
			}
	
			debug("Leave the group for time %f\n", sleep_time);	

			ts.tv_sec = ts.tv_sec + (int)sleep_time;
			ts.tv_nsec = (int)(sleep_time / 1000000000);
			if (ts.tv_nsec > 1000000000) {
				ts.tv_sec++;
				ts.tv_nsec -= 1000000000;
			}
			pthread_cond_timedwait(&dummy_cond, &thread_lock_g, &ts);		

			in_out_flag_g = JOINED;
		}	
	}

	return 0;
}

/*
 * Exit the program when the testing time expires.
 */
static void* exit_handler() {
	pthread_cond_t dummy_cond;              // dummy condition variable
    struct timeval tv;                      // time structures
    struct timespec ts;

	int running_time = RUN_TIME;			// duration of running the program

	// set the timer 
	pthread_cond_init(&dummy_cond, NULL);
	gettimeofday(&tv, NULL);                // get the current time
	ts.tv_sec = tv.tv_sec;
	ts.tv_nsec = tv.tv_usec * 1000;
	
	// wait	
	ts.tv_sec = ts.tv_sec + running_time;	
	pthread_cond_timedwait(&dummy_cond, &thread_lock_g, &ts);		

	// exit the program
	if (in_out_flag_g == JOINED) {
		SEAL_leave(seal_session_g);
	}
	printf("Bye bye: %s\n", seal_session_g->uid);	
	fclose(log_fp_g);
	exit(0);

	// never reach here
	return NULL;
}

/*
 * Stream message out periodically 
 * (demonstrate how to encrypt/decrypt message with the constant change of 
 * the group key)
 */
static void* stream_handler(int period) {
	pthread_cond_t dummy_cond;        // dummy condition variable

	struct timeval tv;                // time structures
	struct timespec ts;

	int msg_id = 0;                   // msg id						

	char sample_msg[5][50] =
	{"I can still see grandma and grandpa",
		"Riding off in their wagon on the way to church",
		"Like always I was slow at getting ready",
		"I can still hear grandma shouting at me",
		"I thought I was slow, but you take the case."};

	// initialize the variables
	pthread_cond_init(&dummy_cond, NULL);
	gettimeofday(&tv, NULL);          // get the current time
	ts.tv_sec = tv.tv_sec;
	ts.tv_nsec = tv.tv_usec * 1000;

	// wait
	ts.tv_sec = ts.tv_sec + period;
	pthread_cond_timedwait(&dummy_cond, &thread_lock_g, &ts);

	// start streaming message out periodically	
	while (1) {
		if (in_out_flag_g == JOINED) {
			// sending message	
			SEAL_send(seal_session_g, sample_msg[msg_id], 
					strlen(sample_msg[msg_id]));
			msg_id = (msg_id + 1) % 5;
		}	

		// set the next running schedule
		ts.tv_sec = ts.tv_sec + period;
		pthread_cond_timedwait(&dummy_cond, &thread_lock_g, &ts);
	}
}

/*
 * Wait for incoming messages
 */
static void* msg_recv_handler() {
	char sender[SEAL_GUID_LEN];
	char msg[SEAL_MAX_MSG_SIZE];

	// set cancel state
	pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
	pthread_setcanceltype(PTHREAD_CANCEL_DEFERRED, NULL);

	// loop to wait for messages    
	while (1) {
		memset(sender, 0, SEAL_GUID_LEN);
		memset(msg, 0, SEAL_MAX_MSG_SIZE);
		if (SEAL_recv(seal_session_g, sender, msg) >= 0) {
			printf("Message from %s: %s\n", sender, msg);
		} else {
			fprintf(stderr, "ERROR: error occurs in msg recv\n");
			break;
		}
	}
	return NULL;
}

/*
 * Print the usage
 */
static void print_usage(char* cmd) {
	printf("Usage: %s [options]\n", cmd);
	printf("Options:\n");
	printf("  [no options]\n");
	printf("      * Joins/leaves randomly at mean 60sec/60sec.\n"); 
	printf("          Records are printed in stdout.\n");
	printf("  -r <in_time> <out_time>\n");
	printf("      * Joins/leaves randomly at specified average times.\n");
	printf("  -p <pattern file> <output file>\n");
	printf("      * Joins/leaves following the specified pattern file.\n");
	printf("  -s <poll_period>\n");
	printf("      * Streams out data periodically.\n");
	printf("  -h\n");
	printf("      * Prints the help menu.\n");
}


/************************************************************************/
/*-----------------------------------------------------------------------
 *					GAUGER IMPLEMENTATION
 *-----------------------------------------------------------------------*/

/*
 * Initialize the gauger
 */
GAUGER_control_p_t GAUGER_init() {
	GAUGER_control_p_t gauger;

	gauger = (GAUGER_control_p_t)malloc(sizeof(GAUGER_control_t));

	gauger->no_exp = 0;
	gauger->no_bcast = 0;
	gauger->no_bkey= 0;
	gauger->is_checking = 0;
	
	return gauger;
}


/*
 * Increment the no_exp count.
 */
void GAUGER_inc_exp(GAUGER_control_p_t gauger) {
	if (gauger->is_checking == 1) {
		gauger->no_exp++;
	}
}

/*
 * Increment the no_bcast count.
 */
void GAUGER_inc_bcast(GAUGER_control_p_t gauger) {
	if (gauger->is_checking == 1) {
		gauger->no_bcast++;
	}
}

/*
 * Increment the no_bkey count.
 */
void GAUGER_inc_bkey(GAUGER_control_p_t gauger, int no_bkey) {
	if (gauger->is_checking == 1) {
		gauger->no_bkey += no_bkey;
	}
}

/*
 * Start testing the time for rekeying
 */
void GAUGER_rekey_start_test(GAUGER_control_p_t gauger) {
	if (gauger->is_checking == 0) {
		gauger->no_exp = 0;	
		gauger->no_bcast = 0;	
		gauger->no_bkey = 0;	
		gettimeofday(&(gauger->rekey_start_time), NULL);
		gauger->is_checking = 1;
	}
}

/*
 * End testing on the time of rekeying
 */
void GAUGER_rekey_end_test(GAUGER_control_p_t gauger, int rekey_seq) {
	struct timeval diff;
	double rekey_time;

	if (gauger->is_checking == 1) {
		gettimeofday(&(gauger->rekey_end_time), NULL);
		diff.tv_sec = (gauger->rekey_end_time).tv_sec - 
			(gauger->rekey_start_time).tv_sec;
		diff.tv_usec = (gauger->rekey_end_time).tv_usec - 
			(gauger->rekey_start_time).tv_usec;
		if (diff.tv_usec < 0) {
			diff.tv_sec--;
			diff.tv_usec += 1000000;
		}	

		rekey_time = diff.tv_sec + diff.tv_usec / 1000000.0;

//		printf("UID: %s, REKEY_SEQ = %d, ", seal_session_g->uid, rekey_seq);
//		printf("DURATION = %f secs, NO_EXP = %d, NO_MSG = %d\n", 
//				rekey_time, gauger->no_exp, gauger->no_bcast);

		fprintf(log_fp_g, "%d\t%f\t%d\t%d\t%d\n", rekey_seq, rekey_time,
				gauger->no_exp, gauger->no_bcast, gauger->no_bkey);
		fflush(log_fp_g);

		gauger->is_checking = 0;
	}
}

/*
 * Reset all test parameters.
 */
void GAUGER_reset(GAUGER_control_p_t gauger) {
	gauger->no_exp = 0;
	gauger->no_bcast = 0;
	gauger->no_bkey = 0;
	gauger->is_checking = 0;
}

