/*
 * packet_engine.c 
 *
 * - define how to send packets to a group
 */

#include "seal.h"

/*------------------------------------------------------------------------
 *                       PACKET ENGINE
 *------------------------------------------------------------------------*/
/*
 * Init the packet engine
 */
PACKET_control_p_t PACKET_init(int mbox, char const* group) {

	PACKET_control_p_t packet_engine;

	// malloc
	packet_engine = (PACKET_control_p_t)malloc(sizeof(PACKET_control_t));

	// initialize the variables
	packet_engine->mbox = mbox;
	memcpy(packet_engine->group, group, SEAL_GROUP_LEN);	

	// initialize the lock
	pthread_mutex_init(&(packet_engine->lock), NULL);

	return packet_engine;
}

/*
 * Send the specified packet. Return the multicast status.
 */
int PACKET_send(PACKET_control_p_t packet_engine, int packet_type, 
		int pkt_len, void const* packet) {
	int ret = -1;

	// send the packet
	switch (packet_type) {
		case PACKET_REKEY:
			debug("Send rekey packet at length  %d.... ,", pkt_len);
			ret = SP_multicast(packet_engine->mbox, AGREED_MESS, 
					packet_engine->group, PACKET_REKEY, pkt_len, 
					(const char*)packet);
			debug("Done rekey\n");
			break;
		case PACKET_BKEY:
			debug("Send blinded key packet at length %d...., ", pkt_len);
			ret = SP_multicast(packet_engine->mbox, AGREED_MESS, 
					packet_engine->group, PACKET_BKEY, pkt_len, 
					(const char*)packet);
			debug("Done blinded key\n");
			break;
		case PACKET_MSG:
			debug("Send msg packet at length %d...., ", pkt_len);
			ret = SP_multicast(packet_engine->mbox, AGREED_MESS, 
					packet_engine->group, PACKET_MSG, pkt_len, 
					(const char*)packet);
			debug("Done msg\n");
			break;
	}
	return ret;
}

/* 
 * destroy
 */
void PACKET_destroy(PACKET_control_p_t packet_engine) {
	free(packet_engine);
}


/*------------------------------------------------------------------------
 *                       PACKET SENDING METHODS
 *------------------------------------------------------------------------*/
/*
 * Create the rekey packet. Return the data length.
 */
int PACKET_create_rekey(int seq, char mode, int no_mnode, int no_join, 
		int no_leave, char* const* mnode_guids, int* tree_node_ids, 
		char* const* join_guids, char* const* leave_guids, 
		PACKET_rekey_p_t ret_pkt) {

	int pkt_len;
	PACKET_rekey_t pkt;
	unsigned int tmp_htonl;		// storing htonl() data
	int i, n;


	// calculate the pkt_len
	pkt_len = 17 + no_mnode * (SEAL_GUID_LEN + 4) + no_join *
		SEAL_GUID_LEN + no_leave * SEAL_GUID_LEN;

	// create the packet
	tmp_htonl = htonl(seq);
	memcpy(pkt.rekey_seq, &tmp_htonl, 4);
	pkt.rekey_mode = mode;
	tmp_htonl = htonl(no_mnode);
	memcpy(pkt.no_mnode, &tmp_htonl, 4);
	tmp_htonl = htonl(no_join);
	memcpy(pkt.no_join, &tmp_htonl, 4);
	tmp_htonl = htonl(no_leave);
	memcpy(pkt.no_leave, &tmp_htonl, 4);
	n = 0;
	for (i=0; i<no_mnode; ++i) {		// mnode guids + tree node ids
		memcpy(pkt.data + n, mnode_guids[i], SEAL_GUID_LEN); 
		tmp_htonl = htonl(tree_node_ids[i]);		
		memcpy(pkt.data + n + SEAL_GUID_LEN, &tmp_htonl, 4);
		n += SEAL_GUID_LEN + 4;
	}
	for (i=0; i<no_join; ++i) {		// join guids
		memcpy(pkt.data + n, join_guids[i], SEAL_GUID_LEN);
		n += SEAL_GUID_LEN;
	}
	for (i=0; i<no_leave; ++i) {		// leave guids
		memcpy(pkt.data + n, leave_guids[i], SEAL_GUID_LEN);
		n += SEAL_GUID_LEN;
	}

	// create the packet
	memcpy(ret_pkt, &pkt, pkt_len);
	return pkt_len;
}


/*
 * Create the packet of the blinded key of the key node. Return the data
 * length.
 */

int PACKET_create_bli_key(int seq, char rekey_mode, int no_bli_key, 
		int const* tree_node_ids, char* const* bcast_keys, 
		PACKET_bli_key_p_t ret_pkt) {

	int pkt_len;
	PACKET_bli_key_t pkt;
	unsigned int tmp_htonl;		// storing htonl() data
	int i, n;

	// calculate the pkt_len
	pkt_len = 9 + no_bli_key * (4 + SESSKEY_LEN);

	// create the packet
	tmp_htonl = htonl(seq);
	memcpy(pkt.rekey_seq, &tmp_htonl, 4);
	pkt.rekey_mode = rekey_mode;
	tmp_htonl = htonl(no_bli_key);
	memcpy(pkt.no_bli_key, &tmp_htonl, 4);

	n = 0;
	for (i=0; i<no_bli_key; ++i) {
		tmp_htonl = htonl(tree_node_ids[i]);
		memcpy(pkt.data + n, &tmp_htonl, 4);
		memcpy(pkt.data + n + 4, bcast_keys[i], SESSKEY_LEN);
		n += 4 + SESSKEY_LEN;
	}

	// create the packet
	memcpy(ret_pkt, &pkt, pkt_len);
	return pkt_len;
}


/*
 * Create the msg packet. Return the data length.
 */
int PACKET_create_msg(int seq, char const* msg, int actual_len,
		PACKET_msg_p_t ret_pkt) {
	
	int pkt_len;
	PACKET_msg_t pkt;
	unsigned int tmp_htonl;		// storing htonl() data

	// calculate the pkt_len
	pkt_len = 4 + actual_len;	// sizeof(seq) + length(msg)

	// create the packet
	tmp_htonl = htonl(seq);
	memcpy(pkt.rekey_seq, &tmp_htonl, 4);
	memcpy(pkt.data, msg, actual_len);

	// create the packet
	memcpy(ret_pkt, &pkt, pkt_len);
	return pkt_len;
}

