/*
 * process_thread.c
 *
 * - Processing messages, for both MEMBERSHIP and REGULAR types.
 */

#include "seal.h"

#ifdef GAUGER
#include "gauger.h"
#endif

static char* guid_g;

static pthread_t * rekey_send_thread_g;
static pthread_t * rekey_poll_thread_g;

static KEYTREE_control_p_t keytree_engine_g;
static LEADER_control_p_t leader_engine_g;
static MEMBER_control_p_t member_engine_g;
static MSGQUEUE_control_p_t msg_queue_g;
static PACKET_control_p_t packet_engine_g;
static PKTQUEUE_control_p_t pkt_queue_g;
static RQUEUE_control_p_t rekey_queue_g;
static SESSKEY_control_p_t sesskey_engine_g;
#ifdef SIGNATURE
static CERTKEY_control_p_t certkey_engine_g;
#endif

static SEAL_session_p_t seal_g;

/*
 * Process incoming packets
 */
void* PROCESS_init(SEAL_session_p_t seal) {

	int data_len;						// data len return (-ve if errors)
	int service_type;					// service type (REG / MEMBERSHIP)
	char sender[SEAL_GUID_LEN];			// sender name  
	int no_guid;						// number of guids
	char guids[SEAL_MAX_GUID][SEAL_GUID_LEN];	// array of guids
	int packet_type;					// packet type (in packet.h)
	int endian_mismatch;				// check if endian matches
	char recv_buf[SEAL_MAX_PKT_SIZE];	// received data

	// set cancel state
	pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
	pthread_setcanceltype(PTHREAD_CANCEL_DEFERRED, NULL);
	pthread_cleanup_push((void*)&PROCESS_cleanup_handler, NULL);

	// set static variables
	guid_g = seal->guid;
	rekey_send_thread_g = seal->rekey_send_thread;
	rekey_poll_thread_g = seal->rekey_poll_thread;
	keytree_engine_g = seal->keytree_engine;
	leader_engine_g = seal->leader_engine;
	member_engine_g = seal->member_engine;
	msg_queue_g = seal->msg_queue;
	packet_engine_g = seal->packet_engine;
	pkt_queue_g = seal->pkt_queue;
	rekey_queue_g = seal->rekey_queue;
	sesskey_engine_g = seal->sesskey_engine;
#ifdef SIGNATURE
	certkey_engine_g = seal->certkey_engine;
#endif
	seal_g = seal;

	while (1) {
		// dequeue a message
		pthread_mutex_lock(&(pkt_queue_g->lock));
		while (PKTQUEUE_get_size(pkt_queue_g) <= 0) {
			pthread_cond_wait(&(pkt_queue_g->cond), &(pkt_queue_g->lock));
		}
		PKTQUEUE_deq(pkt_queue_g, &service_type, sender, &no_guid, 
				guids, &packet_type, &endian_mismatch, recv_buf, &data_len);
		pthread_mutex_unlock(&(pkt_queue_g->lock));

		// distinguish between packets
		if (Is_membership_mess(service_type)) {

			if (Is_reg_memb_mess(service_type)) {
				if (Is_caused_join_mess(service_type)) {
					PROCESS_mem_join(no_guid, guids, recv_buf);

				} else if (Is_caused_leave_mess(service_type)) {
					PROCESS_mem_leave(no_guid, guids, recv_buf);

				} else if (Is_caused_disconnect_mess(service_type)) {
					PROCESS_mem_leave(no_guid, guids, recv_buf);
			
				} else {
					debug_err("Cannot identify mem pkt type\n");
				}
			}

		} else if (Is_regular_mess(service_type)) {
			packet_type = packet_type & 0xff;
			switch (packet_type) {
				case PACKET_REKEY:
					PROCESS_reg_rekey(sender, recv_buf, data_len);
					break;
				case PACKET_BKEY:
					PROCESS_reg_bkey(sender, recv_buf, data_len);
					break;
				case PACKET_MSG:
					PROCESS_reg_msg(sender, recv_buf, data_len);
					break;
				default:
					debug_err("Unable to recognize packet type\n"); 
			}
		} else {
			debug_err("Unknown service type\n");
		}

		// check the cancel point
		pthread_testcancel();
	}

	pthread_cleanup_pop(1);
	return 0;
}


/********************** PROCESS MEMBERSHIP MESSAGE *********************/

/*
 * Processing join message.
 * Add new join members to the join list. Create the current membership list
 * for the newly join members.
 */
void PROCESS_mem_join(int no_curr_guids, char curr_guids[][SEAL_GUID_LEN], 
		char* data) {
	char* join_guid;
	char** ret_guids;
	int* ret_ts;		
	int i;						// loop counter


	// get the newly join member
	join_guid = (char*)malloc(SEAL_GUID_LEN);
	memset(join_guid, 0, SEAL_GUID_LEN);
	memcpy(join_guid, data + SP_get_vs_set_offset_memb_mess(), SEAL_GUID_LEN);

	debug("Received a join membership msg from %s\n", join_guid);

#ifdef SIGNATURE
	// Check if the joined member is a valid member
	pthread_mutex_lock(&(certkey_engine_g->lock));
	if (!CERTKEY_pubkey_is_exist(certkey_engine_g, join_guid)) {
		pthread_mutex_unlock(&(certkey_engine_g->lock));
		free(join_guid);
		return;
	}
	pthread_mutex_unlock(&(certkey_engine_g->lock));
#endif

	// acquire all necessary locks 
	pthread_mutex_lock(&(keytree_engine_g->lock));
	pthread_mutex_lock(&(leader_engine_g->lock));	
	pthread_mutex_lock(&(member_engine_g->lock));
	pthread_mutex_lock(&(rekey_queue_g->lock));

	// curr_memlist: create memlist for newly join, insert a new member
	if (memcmp(guid_g, join_guid, SEAL_GUID_LEN) == 0) {	
		for (i=0; i<no_curr_guids; ++i) {
			// timestamp of insertion = 0 for curr_memlist indicating
			// the members have joined before the new member
			MEMBER_insert(member_engine_g, MEMBER_CURR_LIST, curr_guids[i], 0);
		}
	} else if (!MEMBER_is_empty(member_engine_g, MEMBER_CURR_LIST)) {
		MEMBER_insert(member_engine_g, MEMBER_CURR_LIST, join_guid, time(NULL));
	}

	// insert into join list	
	MEMBER_insert(member_engine_g, MEMBER_JOIN_LIST, join_guid, 0);

	// Check if leader needs to be elected. Start rekey_bcast_thread if the
	// member becomes a leader 
	if (LEADER_is_elect(leader_engine_g, guid_g, join_guid, 
				LEADER_JOIN_EVENT)) {
		ret_guids = (char**)malloc(no_curr_guids * sizeof(char*));	
		for (i=0; i<no_curr_guids; ++i) {
			ret_guids[i] = (char*)malloc(SEAL_GUID_LEN);
		}
		ret_ts = (int*)malloc(no_curr_guids * sizeof(int));		
		
		MEMBER_get_all(member_engine_g, MEMBER_CURR_LIST, ret_guids, ret_ts);
		if (LEADER_elect(leader_engine_g, guid_g, no_curr_guids, ret_guids, 
					ret_ts)) {

			// set myself to be the leader
			LEADER_set_leader(leader_engine_g, guid_g, guid_g);
			
			// start the rekey threads	
			pthread_create(rekey_send_thread_g, NULL, (void*)&REKEY_SEND_init,
					seal_g);
			pthread_create(rekey_poll_thread_g, NULL, (void*)&REKEY_POLL_init,
					seal_g);
		} else {
			// uncertain who is leader
			LEADER_set_leader(leader_engine_g, guid_g, 0);	
		}

		for (i=0; i<no_curr_guids; ++i) {		// free resources
			free(ret_guids[i]);
		}
		free(ret_guids);
		free(ret_ts);

	} else {			// no leader election is leaded

		////////////////////////////////////////////////////////////
		// do only when QUEUE_BATCH is performed
#ifdef QBATCH
		if (LEADER_is_leader(leader_engine_g)) {

			// enqueue a rekey message 
			RQUEUE_enq(rekey_queue_g, REKEY_QSUBTREE, join_guid, NULL);
			pthread_cond_broadcast(&(rekey_queue_g->empty_cond));
		}
#endif
		////////////////////////////////////////////////////////////

	}
	// free all locks	
	pthread_mutex_unlock(&(rekey_queue_g->lock));
	pthread_mutex_unlock(&(member_engine_g->lock));
	pthread_mutex_unlock(&(leader_engine_g->lock));	
	pthread_mutex_unlock(&(keytree_engine_g->lock));
	
	free(join_guid);
}

/*
 * Processing leave message.
 * Add new leave members to the leave list. Create the current membership list
 * for the newly leave members.
 */
void PROCESS_mem_leave(int no_curr_guids, char curr_guids[][SEAL_GUID_LEN], 
		char* data) {
	char leave_guid[SEAL_GUID_LEN];
	char** ret_curr_guids;
	int *ret_curr_ts;
	int i;

	// get the newly leave member
	memset(leave_guid, 0, SEAL_GUID_LEN);
	memcpy(leave_guid, data + SP_get_vs_set_offset_memb_mess(), SEAL_GUID_LEN);
	
	debug("Received a leave membership msg from %s\n", leave_guid);
#ifdef SIGNATURE
	// Check if the joined member is a valid member
	pthread_mutex_lock(&(certkey_engine_g->lock));
	if (!CERTKEY_pubkey_is_exist(certkey_engine_g, leave_guid)) {
		pthread_mutex_unlock(&(certkey_engine_g->lock));
		return;
	}
	pthread_mutex_unlock(&(certkey_engine_g->lock));
#endif

	// acquire the locks
	pthread_mutex_lock(&(keytree_engine_g->lock));
	pthread_mutex_lock(&(leader_engine_g->lock));
	pthread_mutex_lock(&(member_engine_g->lock));
	pthread_mutex_lock(&(rekey_queue_g->lock));	

	// curr_memlist handling
	MEMBER_delete(member_engine_g, MEMBER_CURR_LIST, leave_guid);

	// inset into leave memlist
	MEMBER_insert(member_engine_g, MEMBER_LEAVE_LIST, leave_guid, 0);

	// check if leader needs elected and starts rekey_bcast_thread if the
	// member becomes a leader 
	if (LEADER_is_elect(leader_engine_g, guid_g, leave_guid, 
				LEADER_LEAVE_EVENT)) {

		// get all members from the current memlist
		ret_curr_guids = (char**)malloc(no_curr_guids * sizeof(char*));	
		for (i=0; i<no_curr_guids; ++i) {
			ret_curr_guids[i] = (char*)malloc(SEAL_GUID_LEN);
		}
		ret_curr_ts = (int*)malloc(no_curr_guids * sizeof(int));		
		MEMBER_get_all(member_engine_g, MEMBER_CURR_LIST, ret_curr_guids, 
				ret_curr_ts);
		
		if (LEADER_elect(leader_engine_g, guid_g, no_curr_guids, 
					ret_curr_guids, ret_curr_ts)) {
			
			// set myself to be the leader
			LEADER_set_leader(leader_engine_g, guid_g, guid_g);

			// start the rekey threads
			pthread_create(rekey_send_thread_g, NULL, (void*)&REKEY_SEND_init,
					seal_g);
			pthread_create(rekey_poll_thread_g, NULL, (void*)&REKEY_POLL_init,
					seal_g);
		} else {
			// uncertain who is leader
			LEADER_set_leader(leader_engine_g, guid_g, 0);
		}
		
		for (i=0; i<no_curr_guids; ++i) {		// free resources
			free(ret_curr_guids[i]);
		}
		free(ret_curr_guids);
		free(ret_curr_ts);

	} else {				// no leader election is needed
		
		// RECOVERY protocol: handle leave if rekey is not done
		if (LEADER_is_leader(leader_engine_g)) {
			if (KEYTREE_is_guid_exist(keytree_engine_g, leave_guid)) {
				// leave guid exist in main tree
				if (keytree_engine_g->root != NULL && 
						keytree_engine_g->root->renewed && 
						!SESSKEY_cache_is_exist(sesskey_engine_g, 0)) {

					// send out rekey message 
					// (determine the rekeying version)
#ifdef REBUILD
					RQUEUE_enq(rekey_queue_g, REKEY_REBUILD, NULL, NULL);
#else
#ifdef BATCH
					RQUEUE_enq(rekey_queue_g, REKEY_BATCH, NULL, NULL);
#else
#ifdef QBATCH
					RQUEUE_enq(rekey_queue_g, REKEY_QMERGE, NULL, NULL);
#endif
#endif
#endif

					pthread_cond_broadcast(&(rekey_queue_g->empty_cond));

				}
			} else if (KEYTREE_is_guid_exist(keytree_engine_g->subtree, 
						leave_guid)) {
				// leave guid exist in subtree
				// send out rekey message to handle the special leave
				RQUEUE_enq(rekey_queue_g, REKEY_QSUBTREE, NULL, leave_guid);
				pthread_cond_broadcast(&(rekey_queue_g->empty_cond));
			}
		}
	}

	pthread_mutex_unlock(&(rekey_queue_g->lock));	
	pthread_mutex_unlock(&(member_engine_g->lock));
	pthread_mutex_unlock(&(leader_engine_g->lock));
	pthread_mutex_unlock(&(keytree_engine_g->lock));
}

/********************** PROCESS REGULAR MESSAGE *********************/

/*
 * Processing rekey message
 * Get the keytree and get the join/leave members
 */
void PROCESS_reg_rekey(char const* sender, char const* recv_pkt, int len) {
	PACKET_rekey_t pkt_recv_rekey;
	int rekey_seq;
	char rekey_mode;
	int no_orig_curr, no_join, no_leave;
	char **orig_curr_guids, **join_guids, **leave_guids;
	int *orig_tree_node_ids;	

	KEYTREE_control_p_t tree;
	int *new_keypath_tree_node_ids;
	int *renewed_list;
	int *bcast_list;
	int new_keypath_len;
	int no_renewed;
	int no_tot_bcast = -1;
	int is_rekey;						// check if rekey is necessary
	int is_keytree_orig_empty;			// check if keytree is originally empty

	int bcast_start_tree_node_id = -1;  // starting bcast tree node id
	int no_bcast = -1;                  // no of bcasts in this round
	int* bcast_tree_node_ids;           // the tree node ids to be bcast
	char** bcast_keys;                  // keys to be bcast

	PACKET_bli_key_t pkt_send_bli_key;	// blinded keys to be sent
	int send_len;						// length of the sending data

	int i, j, k, n;						// loop counters

	debug("Received a rekey message from %s\n", sender);

	// copy the received pakcet to the packet structure
	memcpy(&pkt_recv_rekey, recv_pkt, len);

	// extract the fields from the data
	memcpy(&rekey_seq, pkt_recv_rekey.rekey_seq, 4);
	rekey_seq = ntohl(rekey_seq);
	rekey_mode = pkt_recv_rekey.rekey_mode;				
	memcpy(&no_orig_curr, pkt_recv_rekey.no_mnode, 4);
	no_orig_curr = ntohl(no_orig_curr);
	memcpy(&no_join, pkt_recv_rekey.no_join, 4);
	no_join = ntohl(no_join);
	memcpy(&no_leave, pkt_recv_rekey.no_leave, 4);
	no_leave = ntohl(no_leave);

#ifdef SIGNATURE    
	// verify the signature
	pthread_mutex_lock(&(certkey_engine_g->lock));
	if (len > CERTKEY_LEN && CERTKEY_verify(certkey_engine_g, sender, 
				&pkt_recv_rekey, len - CERTKEY_LEN, (char*)&pkt_recv_rekey 
				+ len - CERTKEY_LEN, CERTKEY_LEN)) {
		debug("Verification succeeded\n");
	} else {
		debug_err("Verification failed for sender uid: %s\n", sender);
		pthread_mutex_unlock(&(certkey_engine_g->lock));
		return;
	}
	pthread_mutex_unlock(&(certkey_engine_g->lock));
#endif

	orig_curr_guids = (char**)malloc(no_orig_curr * sizeof(char*));
	for (i=0; i<no_orig_curr; ++i) {
		orig_curr_guids[i] = (char*)malloc(SEAL_GUID_LEN);
	}
	orig_tree_node_ids = (int*)malloc(no_orig_curr * sizeof(int));
	join_guids = (char**)malloc(no_join * sizeof(char*));
	for (i=0; i<no_join; ++i) {
		join_guids[i] = (char*)malloc(SEAL_GUID_LEN);
	}
	leave_guids = (char**)malloc(no_leave * sizeof(char*));
	for (i=0; i<no_leave; ++i) {
		leave_guids[i] = (char*)malloc(SEAL_GUID_LEN);
	}
	n = 0;
	for (i=0; i<no_orig_curr; ++i) {        // curr guids + tree node ids
		memcpy(orig_curr_guids[i], pkt_recv_rekey.data + n, SEAL_GUID_LEN);
		memcpy(&orig_tree_node_ids[i], pkt_recv_rekey.data + n + 
				SEAL_GUID_LEN, 4);
		orig_tree_node_ids[i] = ntohl(orig_tree_node_ids[i]);
		n += SEAL_GUID_LEN + 4;
	}
	for (i=0; i<no_join; ++i) {     // join guids
		memcpy(join_guids[i], pkt_recv_rekey.data + n, SEAL_GUID_LEN);
		n += SEAL_GUID_LEN;
	}
	for (i=0; i<no_leave; ++i) {        // leave guids
		memcpy(leave_guids[i], pkt_recv_rekey.data + n, SEAL_GUID_LEN);
		n += SEAL_GUID_LEN;
	}

	// acquiring the locks
	pthread_mutex_lock(&(keytree_engine_g->lock));
	pthread_mutex_lock(&(leader_engine_g->lock));
	pthread_mutex_lock(&(member_engine_g->lock));
	pthread_mutex_lock(&(msg_queue_g->lock));	
	pthread_mutex_lock(&(packet_engine_g->lock));	
	pthread_mutex_lock(&(rekey_queue_g->lock));	
	pthread_mutex_lock(&(sesskey_engine_g->lock));

	// check the validity of the sender of the rekey message 
	if (!LEADER_is_valid(leader_engine_g, sender)) {
		// FATAL ERROR: a dishonest member claims to be the leader
		debug_err("Dishonest member %s claims to be the leader\n", sender);
		exit(-1);	
	}	
	// for the non-leaders to set the current leader
	if (LEADER_is_changed(leader_engine_g)) {
		LEADER_set_leader(leader_engine_g, guid_g, sender);
	}
	// Synchronize on the sequence number first
	LEADER_seq_max(leader_engine_g, rekey_seq);

	// reset the join/leave list (delete one by one, to avoid that some
	// joins/leaves have not been included in the rekeying message but being
	// removed)
	if (rekey_mode != REKEY_QSUBTREE) {
		for (i=0; i<no_join; ++i) {		// delete one by one for joins
			MEMBER_delete(member_engine_g, MEMBER_JOIN_LIST, join_guids[i]);
		}
		for (i=0; i<no_leave; ++i) {	// delete one by one for leaves
			MEMBER_delete(member_engine_g, MEMBER_LEAVE_LIST, leave_guids[i]);
		}
	} 

    // check for duplicate guid in join and leave (occured when a
    // member joins for a short time) (HANDLED BY REBUILD/BATCH only)
	for (i=0; i<no_join; i++) {
		for (j=0; j<no_leave; ++j) {
			if (memcmp(join_guids[i], leave_guids[j], SEAL_GUID_LEN) == 0 &&
				!KEYTREE_is_guid_exist(keytree_engine_g->subtree, 
					join_guids[i])) {
				// check if leave has not been processed in the subtree
				
				for (k=i; k<no_join-1; ++k) {
					memcpy(join_guids[k], join_guids[k+1], SEAL_GUID_LEN);
				}
				for (k=j; k<no_leave-1; ++k) {
					memcpy(leave_guids[k], leave_guids[k+1], SEAL_GUID_LEN);
				}
				free(join_guids[no_join-1]);
				free(leave_guids[no_leave-1]);
				no_join--; 
				no_leave--;
				i--; 
				break;
			} 
		}
	}

	//---------------------------------------------------------------
	// ***** START: REKEYING *****
	//---------------------------------------------------------------	
	debug("Rekey start: at time %d, rekey_seq(%d), mode: %d, J=%d, L=%d\n",
			time(NULL), rekey_seq, rekey_mode, no_join, no_leave);

	//--------------------------------------------------------------
	// (?) Check if exchanging the keys is necessary first, i.e., my
	// join event has taken effect in the rekey message
	//--------------------------------------------------------------
	is_rekey = 1;
	if (no_join == 0 && no_leave == 0) {
		is_rekey = 0;
	} else if (KEYTREE_is_empty(keytree_engine_g)) {
		if (rekey_mode != REKEY_QSUBTREE) {
			for (i=0; i<no_join; ++i) {
				if (memcmp(join_guids[i], guid_g, SEAL_GUID_LEN) == 0) {
					break;      // joined member included in the join list
				}
			}
			if (i == no_join) {
				is_rekey = 0;
				debug("The join has not taken effect yet in full rekey\n");
				for (i=0; i<no_join; ++i) {
					debug("..%s", join_guids[i]);
				}
				debug("\n");
			}
		} else if (rekey_mode == REKEY_QSUBTREE) {
			for (i=0; i<no_orig_curr; ++i) {
				if (memcmp(orig_curr_guids[i], guid_g, SEAL_GUID_LEN) == 0) {
					break;      // joined member included in the orig subtree
				}
			}
			if (i == no_orig_curr && memcmp(join_guids[0], guid_g,
						SEAL_GUID_LEN) != 0) {
				is_rekey = 0;
				debug("The join has not taken effect yet in int rekey\n");
				debug("..%s\n", join_guids[0]);
			}
		}
	}

	//--------------------------------------------------------------
	// (?) Need rekeying
	//--------------------------------------------------------------
	if (is_rekey) {             // is rekey == 1

#ifdef GAUGER       // START TIMING on REKEYING
		if (rekey_mode != REKEY_QSUBTREE) {
			GAUGER_rekey_start_test(gauger_g);
		}
#endif

		//--------------------------------------------------------------
		// (?) Determine the state before rekeying
		//--------------------------------------------------------------
		is_keytree_orig_empty = KEYTREE_is_empty(keytree_engine_g);

		//--------------------------------------------------------------    
		// cleanup all markings before proceeding
		//--------------------------------------------------------------
		KEYTREE_unmark(keytree_engine_g);

		//--------------------------------------------------------------
		// (?) Determine the tree to work on
		//--------------------------------------------------------------
		if (rekey_mode != REKEY_QSUBTREE) {
			tree = keytree_engine_g;
		} else {
			tree = keytree_engine_g->subtree;
		}

		//--------------------------------------------------------------
		// (?) Create the keytree if empty
		//--------------------------------------------------------------
		if (KEYTREE_is_empty(tree)) {
			KEYTREE_create_from_list(tree, orig_curr_guids, orig_tree_node_ids,
					no_orig_curr);
		}

		//--------------------------------------------------------------
		// (?) Update the keytree based on the rekey mode
		//--------------------------------------------------------------

		// reforming the tree
		if (KEYTREE_is_empty(tree)) {   // no leave is handled on empty tree
			no_leave = 0;
		}
		switch (rekey_mode) {
			case REKEY_REBUILD:
				KEYTREE_rebuild(tree, join_guids, no_join, leave_guids,
						no_leave);
				break;
			case REKEY_BATCH:
				KEYTREE_batch(tree, join_guids, no_join, leave_guids, no_leave);
				break;
			case REKEY_QSUBTREE:
				KEYTREE_queue_subtree(tree, join_guids, no_join, leave_guids,
						no_leave);
				break;
			case REKEY_QMERGE:
				KEYTREE_queue_merge(tree, join_guids, no_join, leave_guids,
						no_leave);
				break;
		}

		//-----------------------------------------------------------------
		// (?) Update the session key engine 
		//------------------------------------------------------------------
		if (!(rekey_mode == REKEY_QSUBTREE && !is_keytree_orig_empty)) {
			// need to update the key engine

			// create a new keypath, with new tree node ids (start from tail)
			new_keypath_len = KEYTREE_get_keypath_len(tree, guid_g);
			new_keypath_tree_node_ids = (int*)malloc(new_keypath_len *
					sizeof(int));
			KEYTREE_get_keypath(tree, guid_g, new_keypath_tree_node_ids);

			// prepare to get renewed nodes and broadcast nodes
			renewed_list = (int*)malloc(new_keypath_len * sizeof(int));
			bcast_list = (int*)malloc(new_keypath_len * sizeof(int));

			// get those of renewed nodes after the keytree is updated
			no_renewed = KEYTREE_get_renewed(tree, guid_g, renewed_list);

			// find the total number of keys to be broadcast by the member
			no_tot_bcast = KEYTREE_sponsor_coordinate(tree, guid_g, bcast_list);

			// update the keypath   
			bcast_start_tree_node_id = SESSKEY_keypath_update(sesskey_engine_g,
					rekey_seq, new_keypath_tree_node_ids, new_keypath_len,
					renewed_list, no_renewed, bcast_list, no_tot_bcast);

			// clean the sesskey_engine->cache if expired
			SESSKEY_cache_delete_expire(sesskey_engine_g, rekey_seq);

			// move any cache key to keypath and compute the key
			SESSKEY_cache_to_keypath(sesskey_engine_g);

			// broadcast the key
			if (bcast_start_tree_node_id > 0) {
				bcast_tree_node_ids = (int*)malloc(new_keypath_len *
						sizeof(int));
				bcast_keys = (char**)malloc(new_keypath_len * sizeof(char*));
				for (i=0; i<new_keypath_len; ++i) {
					bcast_keys[i] = (char*)malloc(SESSKEY_LEN);
				}

				no_bcast = SESSKEY_keypath_retrieve_bcast(sesskey_engine_g,
						bcast_start_tree_node_id, bcast_tree_node_ids,
						bcast_keys);

				// send the blinded keys
				if (no_bcast > 0) {
					memset(&pkt_send_bli_key, 0, sizeof(PACKET_bli_key_t));
					send_len = PACKET_create_bli_key(rekey_seq, rekey_mode, 
							no_bcast, bcast_tree_node_ids, bcast_keys, 
							&pkt_send_bli_key);
#ifdef SIGNATURE
					// sign the packet
					pthread_mutex_lock(&(certkey_engine_g->lock));
					if (CERTKEY_sign(certkey_engine_g, &pkt_send_bli_key, 
								send_len, (char*)(&pkt_send_bli_key) + 
								send_len) == 0) {
						// attach the signature at the end of the packet    
						debug_err("Signature failed\n");
					}
					pthread_mutex_unlock(&(certkey_engine_g->lock));
					send_len += CERTKEY_LEN;
#endif
					PACKET_send(packet_engine_g, PACKET_BKEY, send_len, 
							&pkt_send_bli_key);
				}

				for (i=0; i<new_keypath_len; ++i) {
					free(bcast_keys[i]);
				}
				free(bcast_keys);
				free(bcast_tree_node_ids);
			}

			// Check if the groupkey is formed
			if (SESSKEY_keypath_is_root(sesskey_engine_g) &&
					rekey_mode != REKEY_QSUBTREE) {
				SESSKEY_keypath_to_groupkey(sesskey_engine_g, rekey_seq);

#ifdef GAUGER       // END TIMING on REKEYING
				GAUGER_rekey_end_test(gauger_g, rekey_seq);
#endif
			}

			// free malloc resources    
			free(new_keypath_tree_node_ids);
			free(renewed_list);
			free(bcast_list);

		}       // End: on updating the session key engine
	}   // End: is_rekey == 1

	debug("End processing rekey, no bcast: %d, 1st bcast id: %d\n", no_bcast,
			bcast_start_tree_node_id);
	
	//---------------------------------------------------------------
	// ***** END: REKEYING *****
	//---------------------------------------------------------------	

	// signal any threads that are waiting for the queue
	if (MSGQUEUE_get_size(msg_queue_g) > 0 &&
			SESSKEY_groupkey_is_formed(sesskey_engine_g, 
				MSGQUEUE_get_head_rekey_seq(msg_queue_g))) {
		pthread_cond_broadcast(&(msg_queue_g->cond));
	}
	// end: signal thread

	// remove head of a rekey message (for leader only)
	if (LEADER_is_leader(leader_engine_g)) {
		RQUEUE_set_stopflag(rekey_queue_g, 0);	
		pthread_cond_broadcast(&(rekey_queue_g->stop_cond));
	}

	// release the lock
	pthread_mutex_unlock(&(sesskey_engine_g->lock));
	pthread_mutex_unlock(&(rekey_queue_g->lock));	
	pthread_mutex_unlock(&(keytree_engine_g->lock));
	pthread_mutex_unlock(&(packet_engine_g->lock));	
	pthread_mutex_unlock(&(msg_queue_g->lock));	
	pthread_mutex_unlock(&(member_engine_g->lock));	
	pthread_mutex_unlock(&(leader_engine_g->lock));

	// free all resources
	for (i=0; i<no_orig_curr; ++i)
		free(orig_curr_guids[i]);
	for (i=0; i<no_join; ++i)
		free(join_guids[i]);
	for (i=0; i<no_leave; ++i)
		free(leave_guids[i]);
	free(orig_tree_node_ids);
	free(orig_curr_guids);
	free(join_guids);
	free(leave_guids);
}


/*
 * Processing received bkey
 */
void PROCESS_reg_bkey(char const* sender, char const* recv_pkt, int len) {

	PACKET_bli_key_t pkt_recv_bli_key;	// a blinded key packet
	
	int rekey_seq;						// rekey sequence number
	char rekey_mode;					// rekey mode
	int no_bli_key;						// no of pub keys received
	int* tree_node_ids;					// tree node ids
	char** bli_keys;					// pub keys

	int is_compute = 0;					// check if blinded keys are computed
	int is_cache_update = 0;			// check if the cache is updated
	int is_confirm_need = 0;			// check if confirmation is needed
	int ret_confirm;					// returned value of confirmed result

	int max_comp_tree_node_id = -1;		// max computed tree node id
	int no_bcast = -1;                  // no of bcasts in this round
	int* bcast_tree_node_ids;           // the tree node ids to be bcast
	char** bcast_keys;                  // keys to be bcast
	int keypath_len;					// keypath length

	PACKET_bli_key_t pkt_send_bli_key;	// blinded key packet to be sent
	int send_len;						// length of the sending data

	int i, n;							// loop counters

	// copy the received pakcet to the packet structure
	memcpy(&pkt_recv_bli_key, recv_pkt, len);

	// extract the static fields from the packet
	memcpy(&rekey_seq, pkt_recv_bli_key.rekey_seq, 4);
	rekey_seq = ntohl(rekey_seq);
	rekey_mode = pkt_recv_bli_key.rekey_mode;
	memcpy(&no_bli_key, pkt_recv_bli_key.no_bli_key, 4);
	no_bli_key = ntohl(no_bli_key);

#ifdef SIGNATURE    
	// verify the signature
	pthread_mutex_lock(&(certkey_engine_g->lock));
	if (len > CERTKEY_LEN && CERTKEY_verify(certkey_engine_g, sender, 
				&pkt_recv_bli_key, len - CERTKEY_LEN, (char*)&pkt_recv_bli_key 
				+ len - CERTKEY_LEN, CERTKEY_LEN)) {
		debug("Verification succeeded\n");
	} else {
		debug_err("Verification failed for sender uid: %s\n", sender);
		pthread_mutex_unlock(&(certkey_engine_g->lock));
		return;
	}
	pthread_mutex_unlock(&(certkey_engine_g->lock));
#endif
	
	// no processing for existing members if QUEUE_SUBTREE is performed
	pthread_mutex_lock(&(keytree_engine_g->lock));
	if ((rekey_mode == REKEY_QSUBTREE && !KEYTREE_is_empty(keytree_engine_g)) ||
		(rekey_mode != REKEY_QSUBTREE && KEYTREE_is_empty(keytree_engine_g))) {
		pthread_mutex_unlock(&(keytree_engine_g->lock));
		return;
	}
	pthread_mutex_unlock(&(keytree_engine_g->lock));

	// extract the variable field from the packet
	tree_node_ids = (int*)malloc(no_bli_key * sizeof(int));
	bli_keys = (char**)malloc(no_bli_key * sizeof(char*));
	for (i=0; i<no_bli_key; ++i) {
		bli_keys[i] = (char*)malloc(SESSKEY_LEN);
	}	
	n = 0;
	for (i=0; i<no_bli_key; ++i) {
		memcpy(&tree_node_ids[i], pkt_recv_bli_key.data + n, 4);
		tree_node_ids[i] = ntohl(tree_node_ids[i]);
		memcpy(bli_keys[i], pkt_recv_bli_key.data + n + 4, SESSKEY_LEN);
		n += 4 + SESSKEY_LEN;
	}	
	
	debug("Received pub key from %s seq %d with %d keys, 1st key: %d\n", 
			sender, rekey_seq, no_bli_key, tree_node_ids[0]);

#ifdef GAUGER		// INCREMENT THE NO_BCAST COUNT
	GAUGER_inc_bcast(gauger_g);
	GAUGER_inc_bkey(gauger_g, no_bli_key);
#endif

	// acquire the locks
	pthread_mutex_lock(&(msg_queue_g->lock));
	pthread_mutex_lock(&(packet_engine_g->lock));
	pthread_mutex_lock(&(sesskey_engine_g->lock));

	// check if the blinded key is present, or compute key if necessary
	for (i=0; i<no_bli_key; ++i) {
	
		is_cache_update = SESSKEY_cache_update(sesskey_engine_g, 
				tree_node_ids[i], bli_keys[i], rekey_seq);
		
		if (is_cache_update) {

			// try to compute keys up the keypath based on pub keys received
			if (SESSKEY_keypath_compute(sesskey_engine_g, tree_node_ids[i], 
					bli_keys[i], rekey_seq) > 0) {
			
				is_compute = 1;		// set to 1 if keypath gets computed
				
				if ((tree_node_ids[i] - 1) / 2 > max_comp_tree_node_id) {
					max_comp_tree_node_id = (tree_node_ids[i] - 1) / 2;	
				}
			}	
			
			// reset the bcast flag of the pub key 
			SESSKEY_keypath_unmark_bcast(sesskey_engine_g, tree_node_ids[i]);

			// check if a new root key is received
			if (tree_node_ids[i] == 0) {
				is_confirm_need = 1;		// confirmation is needed
			}
		}
	}

	// broadcast new blinded keys from newly computed nodes
	if (is_compute) {
		keypath_len = SESSKEY_keypath_get_len(sesskey_engine_g);
		bcast_tree_node_ids = (int*)malloc(keypath_len * sizeof(int));
		bcast_keys = (char**)malloc(keypath_len * sizeof(char*));
		for (i=0; i<keypath_len; ++i) {
			bcast_keys[i] = (char*)malloc(SESSKEY_LEN);
		}

		no_bcast = SESSKEY_keypath_retrieve_bcast(sesskey_engine_g,
				max_comp_tree_node_id, bcast_tree_node_ids,
				bcast_keys);

		if (no_bcast > 0) {
			memset(&pkt_send_bli_key, 0, sizeof(PACKET_bli_key_t));
			send_len = PACKET_create_bli_key(rekey_seq, rekey_mode, 
					no_bcast, bcast_tree_node_ids, bcast_keys, 
					&pkt_send_bli_key);
#ifdef SIGNATURE
			// sign the packet
			pthread_mutex_lock(&(certkey_engine_g->lock));
			if (CERTKEY_sign(certkey_engine_g, &pkt_send_bli_key, 
						send_len, (char*)(&pkt_send_bli_key) + 
						send_len) == 0) {
				// attach the signature at the end of the packet    
				debug_err("Signature failed\n");
			}
			pthread_mutex_unlock(&(certkey_engine_g->lock));
			send_len += CERTKEY_LEN;
#endif
			PACKET_send(packet_engine_g, PACKET_BKEY, send_len, 
					&pkt_send_bli_key);
		}

		for (i=0; i<keypath_len; ++i) {
			free(bcast_keys[i]);
		}
		free(bcast_keys);
		free(bcast_tree_node_ids);
	}

	// change the state if possible
	if (is_compute && !SESSKEY_groupkey_is_formed(sesskey_engine_g, rekey_seq) 
			&& SESSKEY_keypath_is_root(sesskey_engine_g)) {
		if (rekey_mode != REKEY_QSUBTREE) {
			SESSKEY_keypath_to_groupkey(sesskey_engine_g, rekey_seq);
		
			is_confirm_need = 1;	
		} else {
			debug("I got a tmp root key for queue subtree\n");	
		}
	}
	
	// key confirmation -- check if received == computed
	// The protocol is invoked if cache is updated && the blinded group key 
	// is received, OR group key is computed
	if (rekey_mode != REKEY_QSUBTREE && is_confirm_need) {

		// confirmation (when the blinded key is received)
		ret_confirm = SESSKEY_confirm(sesskey_engine_g, rekey_seq);
		if (ret_confirm > 0) {
			debug("Confirmed group key at time %d with seq: %d\n", 
					(int)time(NULL), rekey_seq);

#ifdef GAUGER		// END TIMING on REKEYING
			GAUGER_rekey_end_test(gauger_g, rekey_seq);
#endif

			// signal any threads that are waiting for the queue
			if (MSGQUEUE_get_size(msg_queue_g) > 0 &&
				SESSKEY_groupkey_is_formed(sesskey_engine_g, 
					MSGQUEUE_get_head_rekey_seq(msg_queue_g))) {
				pthread_cond_broadcast(&(msg_queue_g->cond));
			}
			// end: signal thread

		} else if (ret_confirm < 0) {
			// FATAL ERROR: key confirmation fails	
			debug_err("Group key is not confirmed with seq: %d\n",
					rekey_seq);
			SESSKEY_keypath_print(sesskey_engine_g);
			exit(-1);
		}
	}

	// release the locks
	pthread_mutex_unlock(&(sesskey_engine_g->lock));
	pthread_mutex_unlock(&(packet_engine_g->lock));
	pthread_mutex_unlock(&(msg_queue_g->lock));

	// free resources
	for (i=0; i<no_bli_key; ++i) {
		free(bli_keys[i]);
	}
	free(bli_keys);
	free(tree_node_ids);
}	

/*
 * Processing message data.
 * Enqueue the message into the message queue.
 */
void PROCESS_reg_msg(char const* sender, char const* recv_pkt, int len) {
	PACKET_msg_t pkt;						// a msg packet
	int rekey_seq;							// key's rekey sequence number
	char encrypt_msg[SEAL_MAX_MSG_SIZE];	// encrypted message data
											//	Encrypt[plain msg + signature]
	int encrypt_len = -1;
	
	debug("Received a msg packet from req_guid: %s\n", sender);

	// copy the received pakcet to the packet structure
	memcpy(&pkt, recv_pkt, len);

	// extract the static fields from the packet
	memcpy(&rekey_seq, pkt.rekey_seq, 4);
	rekey_seq = ntohl(rekey_seq);

	// extract the variable field from the packet	
	memset(encrypt_msg, 0, SEAL_MAX_MSG_SIZE);
	encrypt_len = len - 4;
	memcpy(encrypt_msg, pkt.data, encrypt_len);

	// acquire the locks
	pthread_mutex_lock(&(msg_queue_g->lock));
	pthread_mutex_lock(&(sesskey_engine_g->lock));

	// enqueue the encrypted message (only when the corresponding group key 
	// is formed)
	if (SESSKEY_groupkey_is_formed(sesskey_engine_g, rekey_seq)) {
		MSGQUEUE_enq(msg_queue_g, rekey_seq, sender, encrypt_msg, encrypt_len);
		pthread_cond_broadcast(&(msg_queue_g->cond));
	}

	// release the locks
	pthread_mutex_unlock(&(sesskey_engine_g->lock));
	pthread_mutex_unlock(&(msg_queue_g->lock));

}

/*********** PROCESS cleanup handler ***************/
void* PROCESS_cleanup_handler() {
	if (*rekey_poll_thread_g != 0) {
		pthread_cancel(*rekey_poll_thread_g);
	}
	if (*rekey_send_thread_g != 0) {
		pthread_cancel(*rekey_send_thread_g);
	}

#ifdef SIGNATURE
	pthread_mutex_trylock(&(certkey_engine_g->lock));
#endif

	pthread_mutex_trylock(&(keytree_engine_g->lock));
	pthread_mutex_trylock(&(leader_engine_g->lock));	
	pthread_mutex_trylock(&(member_engine_g->lock));      
	pthread_mutex_trylock(&(msg_queue_g->lock));
	pthread_mutex_trylock(&(packet_engine_g->lock));
	pthread_mutex_trylock(&(pkt_queue_g->lock));
	pthread_mutex_trylock(&(rekey_queue_g->lock));
	pthread_mutex_trylock(&(sesskey_engine_g->lock));
	
	pthread_mutex_unlock(&(sesskey_engine_g->lock));
	pthread_mutex_unlock(&(rekey_queue_g->lock));
	pthread_mutex_unlock(&(pkt_queue_g->lock));
	pthread_mutex_unlock(&(packet_engine_g->lock));
	pthread_mutex_unlock(&(msg_queue_g->lock));
	pthread_mutex_unlock(&(member_engine_g->lock));      
	pthread_mutex_unlock(&(leader_engine_g->lock));
	pthread_mutex_unlock(&(keytree_engine_g->lock));

#ifdef SIGNATURE
	pthread_mutex_unlock(&(certkey_engine_g->lock));
#endif
	debug("PROCESS->cleanup\n"); 
	return NULL;
}

