/*
 * seal.h
 */

/* 
 * include the libraries
 */
#include <ctype.h>
#include <errno.h>
#include <netdb.h>
#include <pthread.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <termios.h>
#include <time.h>
#include <unistd.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <openssl/bn.h>
#include <openssl/des.h>
#include <openssl/dh.h>
#include <openssl/evp.h>
#include <openssl/md5.h>
#include <openssl/pem.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/time.h>

#ifdef GUI_ENABLED
#include <gtk/gtk.h>
#endif

/* common constants */
#define SEAL_UID_LEN		16			// * length of user id
#define SEAL_GUID_LEN		32          // * length of GUID 
										//   (equals MAX_GROUP_NAME in Spread)
#define SEAL_GROUP_LEN		32			// * length of group name	
#define SEAL_IP_LEN			20          // * length of ipv4 address
#define SEAL_MAX_GUID		100			// * maximum number of guids
										//   (used in for SP_receive() and
										//   packet queue, dynamic allocation 
										//   is encouraged to reduce memory)
#define SEAL_MAX_PKT_SIZE	8000        // * max packet size
#define SEAL_MAX_MSG_SIZE	3000		// * max message size 
										//   (including signature if present)
#define SEAL_PASSWD_LEN		8			// * passwd size (signature mode only)

#define SEAL_REBUILD		0x01
#define SEAL_BATCH			0x02
#define SEAL_QSUBTREE		0x03
#define SEAL_QMERGE			0x04

#ifndef __SEAL_INCLUDE			// ifndef __SEAL_INCLUDE
#define __SEAL_INCLUDE

/* forward declaration */
typedef struct seal_session_s * SEAL_session_p_t;

/* header files */
#include "certkey_engine.h"
#include "config.h"
#include "debug.h"
#include "keytree_engine.h"
#include "leader_engine.h"
#include "member_engine.h"
#include "message_queue.h"
#include "packet_engine.h"
#include "packet_queue.h"
#include "process_thread.h"
#include "receive_thread.h"
#include "rekey_poll_thread.h"
#include "rekey_queue.h"
#include "rekey_send_thread.h"
#include "sesskey_engine.h"
#include "sp.h"

/* global variables */
typedef struct seal_session_s {
	char uid[SEAL_UID_LEN];				// unique id
	char spread_ip[SEAL_IP_LEN];		// spread ip
	int spread_port;					// spread port
	int salt;							// salt (attached after uid_g)

	char passwd[SEAL_PASSWD_LEN + 1];	// password

	char group[SEAL_GROUP_LEN];			// group name
	char guid[SEAL_GUID_LEN];			// globally unique id
	int mbox;							// message box (socket)

	pthread_t * receive_thread;
	pthread_t * process_thread;
	pthread_t * rekey_send_thread;
	pthread_t * rekey_poll_thread;

	KEYTREE_control_p_t keytree_engine;
	LEADER_control_p_t leader_engine;
	MEMBER_control_p_t member_engine;
	MSGQUEUE_control_p_t msg_queue;
	PACKET_control_p_t packet_engine;
	PKTQUEUE_control_p_t pkt_queue;
	RQUEUE_control_p_t rekey_queue;
	SESSKEY_control_p_t sesskey_engine;
#ifdef SIGNATURE
	CERTKEY_control_p_t certkey_engine;
#endif
} SEAL_session_t;

/* function prototypes */
SEAL_session_p_t SEAL_init();
int SEAL_set_passwd(SEAL_session_p_t seal, char* passwd);
int SEAL_join(SEAL_session_p_t seal, char const* group);
int SEAL_leave(SEAL_session_p_t seal);
int SEAL_recv(SEAL_session_p_t seal, char* sender, void* msg);		
int SEAL_send(SEAL_session_p_t seal, const void* msg, int msg_len);
void SEAL_read_membership(SEAL_session_p_t seal, char** curr_guids, 
		int* no_curr, char** join_guids, int* no_join, char** leave_guids, 
		int* no_leave);
void SEAL_destroy(SEAL_session_p_t seal);

#endif		// endif __SEAL_INCLUDE
