/*
 * sesskey_engine.c 
 *
 * - Responsible for generating and updating the session keys along the key
 * path
 */

#include "seal.h"

#ifdef GAUGER
#include "gauger.h"
#endif

/*
 * Initialize the sesskey_engine
 */
SESSKEY_control_p_t SESSKEY_init() {
	SESSKEY_control_p_t sesskey_engine;
	FILE* fp;
	DH* dh;

	// malloc the key engine
	sesskey_engine = (SESSKEY_control_p_t)malloc(sizeof(SESSKEY_control_t));
	if (sesskey_engine == NULL) {
		debug_err("Unable to create key engine\n");
		return NULL;
	}	
	
	// open the pem file and get the dh parameters
	fp = fopen(SESSKEY_FILE, "r");
	if (fp == NULL) {
		debug_err("Unable to open %s for DH params\n", SESSKEY_FILE);
		return NULL;
	}
	if ((dh = PEM_read_DHparams(fp, NULL, NULL, NULL)) == NULL) {
		debug_err("Unable to read parameters\n");
		return NULL;
	}
	if (!DH_generate_key(dh)) {
		debug_err("Unable to generate session blinded/secret keys\n");
		return NULL;
	}
	fclose(fp);

	// transform the dh parameters to the SESSKEY->param structure
	sesskey_engine->param = (SESSKEY_param_p_t)malloc(sizeof(SESSKEY_param_t));
	memset(sesskey_engine->param, 0, sizeof(SESSKEY_param_t));
	BN_bn2bin(dh->priv_key, sesskey_engine->param->sess_sec_key + SESSKEY_LEN 
			- BN_num_bytes(dh->priv_key));
	BN_bn2bin(dh->pub_key, sesskey_engine->param->sess_bli_key + SESSKEY_LEN
			- BN_num_bytes(dh->pub_key));
	BN_bn2bin(dh->g, sesskey_engine->param->g + SESSKEY_LEN 
			- BN_num_bytes(dh->g));	
	BN_bn2bin(dh->p, sesskey_engine->param->p + SESSKEY_LEN 
			- BN_num_bytes(dh->p));	

	// free the dh structure
	DH_free(dh);	

	// initialize the keypath
	sesskey_engine->keypath = SESSKEY_keypath_create(0);

	// initialize the cache
	sesskey_engine->cache = (SESSKEY_cache_p_t)malloc(sizeof(SESSKEY_cache_t));	
	sesskey_engine->cache->head = NULL;
	sesskey_engine->cache->tail = NULL;
	
	// initialize the group key list
	sesskey_engine->groupkeys = (SESSKEY_groupkey_p_t)
							malloc(sizeof(SESSKEY_groupkey_t));	
	sesskey_engine->groupkeys->size = 0;
	sesskey_engine->groupkeys->head = NULL;
	sesskey_engine->groupkeys->tail = NULL;

	// initialize the lock	
	pthread_mutex_init(&(sesskey_engine->lock), NULL);
	
	return sesskey_engine;
}


/*------------------------------------------------------------------------
 *                            KEYPATH MANAGEMENT
 *------------------------------------------------------------------------*/
/*
 * Create a keypath given the length (with all the entries undefined)
 */
SESSKEY_keypath_p_t SESSKEY_keypath_create(int len) {
	SESSKEY_keypath_p_t keypath;
	SESSKEY_keypath_entry_p_t entry;
	int i;

	keypath = (SESSKEY_keypath_p_t)malloc(sizeof(SESSKEY_keypath_t));	
	keypath->head = NULL;
	keypath->tail = NULL;
	keypath->rekey_seq = -1;

	for (i=0; i<len; ++i) {
		entry = (SESSKEY_keypath_entry_p_t)
			malloc(sizeof(SESSKEY_keypath_entry_t));
		entry->tree_node_id = -1;
		entry->sib_tree_node_id = -1;
		memset(entry->sec_key, 0, SESSKEY_LEN);	
		memset(entry->sib_bli_key, 0, SESSKEY_LEN);
		entry->bcast_flag = 0;
		entry->sec_key_flag = 0;
		entry->bli_key_flag = 0;
		entry->prev = keypath->tail;
		entry->next = NULL;
		if (keypath->head == NULL) {
			keypath->head = entry;
		} else {
			keypath->tail->next = entry;
		}
		keypath->tail = entry;	
	}

	return keypath;	
}


/*
 * Update the keypath of the key engine
 * Returns the tree_node_id of the key to be broadcast
 */
int SESSKEY_keypath_update(SESSKEY_control_p_t sesskey_engine, int rekey_seq,
		int const* new_keypath_tree_node_ids, int new_keypath_len, 
		int const* renewed_list, int no_renewed, int const* bcast_list, 
		int no_bcast) {
	
	SESSKEY_keypath_p_t new_keypath;
	SESSKEY_keypath_entry_p_t old_entry_p, new_entry_p, tmp_entry_p;
	int old_keypath_len;	// original keypath length before update
	int bcast_count;		// check how many to be broadcast
	int renewed_count;		// check how many to be renewed	
	int i;

	// get the old_keypath_len
	old_keypath_len = 0;
	old_entry_p = sesskey_engine->keypath->head;
	while (old_entry_p != NULL) {
		old_keypath_len++;
		old_entry_p = old_entry_p->next;
	}

	// create a new keypath given the keypath length.
	new_keypath = SESSKEY_keypath_create(new_keypath_len);

	// update the keypath's sequence number 
	new_keypath->rekey_seq = rekey_seq; 

	// update tree_node_ids and sib_tree_node_ids in new keypath
	new_entry_p = new_keypath->tail;
	for (i=0; i<new_keypath_len; ++i) {
		new_entry_p->tree_node_id = new_keypath_tree_node_ids[i];
		if (new_entry_p->tree_node_id % 2 == 0) {
			new_entry_p->sib_tree_node_id = new_entry_p->tree_node_id - 1;
		} else {
			new_entry_p->sib_tree_node_id = new_entry_p->tree_node_id + 1;
		}
		new_entry_p = new_entry_p->prev;
	}

	// update secret keys of non-renewed blinded keys in new keypath
	if (old_keypath_len > 0 && old_keypath_len >= new_keypath_len
			- no_renewed) {
		// copy the fields from old to new keypath 
		old_entry_p = sesskey_engine->keypath->tail;
		new_entry_p = new_keypath->tail;
		for (i=0; i < new_keypath_len - no_renewed - 1; ++i) {
			memcpy(new_entry_p->sec_key, old_entry_p->sec_key, SESSKEY_LEN);
			memcpy(new_entry_p->sib_bli_key, old_entry_p->sib_bli_key,
					SESSKEY_LEN);
			new_entry_p->bcast_flag = old_entry_p->bcast_flag;
			new_entry_p->sec_key_flag = old_entry_p->sec_key_flag;
			new_entry_p->bli_key_flag = old_entry_p->bli_key_flag;

			old_entry_p = old_entry_p->prev;
			new_entry_p = new_entry_p->prev;
		}
		memcpy(new_entry_p->sec_key, old_entry_p->sec_key, SESSKEY_LEN);
		new_entry_p->bcast_flag = old_entry_p->bcast_flag;
		new_entry_p->sec_key_flag = old_entry_p->sec_key_flag;
		new_entry_p->bli_key_flag = old_entry_p->bli_key_flag;

	} else if (old_keypath_len == 0) {
		// update itself's secret key to the end of keypath	
		new_entry_p = new_keypath->tail;
		memcpy(new_entry_p->sec_key, sesskey_engine->param->sess_sec_key, 
				SESSKEY_LEN);
	}

	// check if the sec_key and sib_bli_key need to be updated
	tmp_entry_p = new_keypath->tail;
	renewed_count = 0;
	while (tmp_entry_p != NULL && renewed_count < no_renewed) {
		if (renewed_list[renewed_count] == tmp_entry_p->tree_node_id) {
			tmp_entry_p->sec_key_flag = 1;			// needs update
			tmp_entry_p->next->bli_key_flag = 1;	// needs update
			renewed_count++;
		}
		tmp_entry_p = tmp_entry_p->prev;
	}

	// check if the sec_key needs to be bcast
	tmp_entry_p = new_keypath->tail;
	bcast_count = 0;
	while (tmp_entry_p != NULL && bcast_count < no_bcast) {
		if (bcast_list[bcast_count] == tmp_entry_p->tree_node_id) {
			tmp_entry_p->bcast_flag = 1;		// needs bcast
			bcast_count++;
		} 
		tmp_entry_p = tmp_entry_p->prev;
	}
	
	// free the original keypath and replace it with a new one
	tmp_entry_p = sesskey_engine->keypath->head;
	while (tmp_entry_p != NULL) {
		sesskey_engine->keypath->head = tmp_entry_p->next;
		free(tmp_entry_p);
		tmp_entry_p = sesskey_engine->keypath->head;
	}
	free(sesskey_engine->keypath);	
	sesskey_engine->keypath = new_keypath;

	// return the first key to be broadcast
	tmp_entry_p = new_keypath->tail;
	while (tmp_entry_p != NULL) {
		if (tmp_entry_p->bcast_flag == 1) {
			return tmp_entry_p->tree_node_id;
		}
		tmp_entry_p = tmp_entry_p->prev;
	}
	return -1;
}


/*
 * Check if the blinded key exists in the keypath
 * Compute the keys.
 * Return 1 on computing, 0 on no-change
 */
int SESSKEY_keypath_compute(SESSKEY_control_p_t sesskey_engine, 
		int sib_tree_node_id, char const* bli_key, int seq) {
	
	SESSKEY_keypath_entry_p_t tmp_entry_p;

	// find if the blinded key has been updated
	tmp_entry_p = sesskey_engine->keypath->head;
	while (tmp_entry_p != NULL) {
		if (tmp_entry_p->sib_tree_node_id == sib_tree_node_id) {
			memcpy(tmp_entry_p->sib_bli_key, bli_key, SESSKEY_LEN);
			tmp_entry_p->bli_key_flag = 0;
			break;	
		} else if (tmp_entry_p->sib_tree_node_id > sib_tree_node_id) {
			return 0;
		} else {
			tmp_entry_p = tmp_entry_p->next;	
		}
	}

	if (tmp_entry_p == NULL) {		// not found in the keypath
		return 0;
	}

	// update the keys
	while (tmp_entry_p->prev != NULL && tmp_entry_p->sec_key_flag == 0
			&& tmp_entry_p->bli_key_flag == 0) {

		SESSKEY_compute_secret(sesskey_engine, tmp_entry_p->prev->sec_key,
				tmp_entry_p->sec_key, tmp_entry_p->sib_bli_key);

		tmp_entry_p->prev->sec_key_flag = 0;
		tmp_entry_p = tmp_entry_p->prev;
	}

	return 1;
}


/*
 * Retrieve the blinded keys to be broadcast.
 * Return no of bcast keys
 */
int SESSKEY_keypath_retrieve_bcast(SESSKEY_control_p_t sesskey_engine, 
		int start_tree_node_id, int* ret_tree_node_ids, 
		char** ret_bcast_keys) {
			
	SESSKEY_keypath_entry_p_t tmp_entry_p = NULL, start_entry_p = NULL;
	int no_bcast = 0;
	int* tree_node_ids;
	char** bcast_keys;
	int i;

	// go to the position where the tree node may be broadcast
	tmp_entry_p = sesskey_engine->keypath->head;
	while (tmp_entry_p != NULL) {
		if (tmp_entry_p->tree_node_id == start_tree_node_id) {
			break;
		} else {
			tmp_entry_p = tmp_entry_p->next;
		}
	}
	if (tmp_entry_p == NULL) {		// tmp_entry_p should not be null
		debug_err("unexpected error in SESSKEY keypath bcast\n");
	}

	// scan through the keypath to see if any keys can be bcast
	// (tmp_entry_p is at the position where tree_node_id == specified one)
	while (tmp_entry_p != NULL) {
		if (tmp_entry_p->sec_key_flag == 0 && tmp_entry_p->bcast_flag == 1) {
			if (no_bcast == 0) {
				start_entry_p = tmp_entry_p;
			}
			no_bcast++;
		}
		tmp_entry_p = tmp_entry_p->prev;
	}
	
	// bcast if no_bcast > 0
	if (no_bcast > 0) {
		
		// malloc
		tree_node_ids = (int*)malloc(no_bcast * sizeof(int));
		bcast_keys = (char**)malloc(no_bcast * sizeof(char*));
		for (i=0; i<no_bcast; ++i) {
			bcast_keys[i] = (char*)malloc(SESSKEY_LEN);
		}

		// start bcast
		tmp_entry_p = start_entry_p;
		i = 0;
		while (tmp_entry_p != NULL) {	
			if (tmp_entry_p->sec_key_flag == 0 && 
					tmp_entry_p->bcast_flag == 1) {
		
				// get the tree node id	
				tree_node_ids[i] = tmp_entry_p->tree_node_id;

				// calculate the blinded key	
				SESSKEY_compute_blinded(sesskey_engine, bcast_keys[i], 
						tmp_entry_p->sec_key);

				i++;
			}

			// next	
			tmp_entry_p = tmp_entry_p->prev;
		}	
		
		// return the bcast keys and tree node ids
		for (i=0; i<no_bcast; ++i) {	
			ret_tree_node_ids[i] = tree_node_ids[i];
			memcpy(ret_bcast_keys[i], bcast_keys[i], SESSKEY_LEN);
		}
	
		// free resources
		for (i=0; i<no_bcast; ++i) {
			free(bcast_keys[i]);
		}
		free(bcast_keys);
		free(tree_node_ids);

	}	// end: no_bcast > 0

	return no_bcast;
}


/*
 * Reset the broadcast flag upon receive the broadcast key
 */
void SESSKEY_keypath_unmark_bcast(SESSKEY_control_p_t sesskey_engine, 
		int tree_node_id) {
	
	SESSKEY_keypath_entry_p_t tmp_entry_p;
	
	tmp_entry_p = sesskey_engine->keypath->head;
	while (tmp_entry_p != NULL) {
		if (tmp_entry_p->tree_node_id == tree_node_id) {
			tmp_entry_p->bcast_flag = 0;
			break;
		}	
		tmp_entry_p = tmp_entry_p->next;
	}

}

/*
 * copy the group key from keypath to groupkeys
 */
void SESSKEY_keypath_to_groupkey(SESSKEY_control_p_t sesskey_engine, int seq) {
	if (sesskey_engine->keypath->head->sec_key_flag == 0) {
		SESSKEY_groupkey_insert(sesskey_engine, 
				sesskey_engine->keypath->head->sec_key, seq);
	}
}

/*
 * Get the length of the keypath in the session key engine.
 */
int SESSKEY_keypath_get_len(SESSKEY_control_p_t sesskey_engine) {
	
	SESSKEY_keypath_entry_p_t tmp_entry_p;
	int len = 0;
	
	tmp_entry_p = sesskey_engine->keypath->head;
	while (tmp_entry_p != NULL) {
		len++;	
		tmp_entry_p = tmp_entry_p->next;
	}
	return len;
}


/*
 * Check if the group key is computed (for REKEY STATE only)
 * Return 1 on computed, 0 on not computed
 */
int SESSKEY_keypath_is_root(SESSKEY_control_p_t sesskey_engine) {

	if (sesskey_engine->keypath->head != NULL && 
			sesskey_engine->keypath->head->sec_key_flag == 0) {
		return 1;
	}

	return 0;
}


/*
 * Print the keypath
 */
void SESSKEY_keypath_print(SESSKEY_control_p_t sesskey_engine) {
	SESSKEY_keypath_entry_p_t tmp_entry_p;
	int i;

	tmp_entry_p = sesskey_engine->keypath->tail;
	while (tmp_entry_p != NULL) {
		printf("\n=========================================\n");
		printf("%d(%d%d%d)=", tmp_entry_p->tree_node_id, 
				tmp_entry_p->bcast_flag, tmp_entry_p->sec_key_flag, 
				tmp_entry_p->bli_key_flag);
		for (i=0; i<SESSKEY_LEN; ++i) {
			printf("%d,", tmp_entry_p->sec_key[i]);
		}
		tmp_entry_p = tmp_entry_p->prev;
	}
	printf("\n=========================================\n");
}


/*------------------------------------------------------------------------
 *                            CACHE MANAGEMENT
 *------------------------------------------------------------------------*/
/*
 * Insert an element to the cache
 */
void SESSKEY_cache_insert(SESSKEY_control_p_t sesskey_engine, int tree_node_id,
		char const* bli_key, int seq) {
	SESSKEY_cache_entry_p_t new_entry_p;

	new_entry_p = (SESSKEY_cache_entry_p_t)malloc(sizeof(
				SESSKEY_cache_entry_t));
	if (new_entry_p == NULL) {
		debug_err("Unable to create a new keyeng cache entry\n");
		return;
	}

	memset(new_entry_p, 0, sizeof(SESSKEY_cache_entry_t));
	new_entry_p->tree_node_id = tree_node_id;
	memcpy(new_entry_p->bli_key, bli_key, SESSKEY_LEN);
	new_entry_p->rekey_seq = seq;

	new_entry_p->prev = sesskey_engine->cache->tail;
	new_entry_p->next = NULL;
	if (sesskey_engine->cache->head == NULL) {
		sesskey_engine->cache->head = new_entry_p;
	} else {
		sesskey_engine->cache->tail->next = new_entry_p;
	}
	sesskey_engine->cache->tail = new_entry_p;
}

/* 
 * Delete any expired elements in the cache, but keep the cache's control
 */
void SESSKEY_cache_delete_expire(SESSKEY_control_p_t sesskey_engine, 
		int rekey_seq) {
	SESSKEY_cache_entry_p_t tmp_entry_p, del_entry_p;

	// remove expired entries in the list
	tmp_entry_p = sesskey_engine->cache->head;
	while (tmp_entry_p != NULL) {
		if (tmp_entry_p->rekey_seq < rekey_seq) {
			del_entry_p = tmp_entry_p;
			tmp_entry_p = tmp_entry_p->next;

			if (del_entry_p->prev != NULL) {
				del_entry_p->prev->next = del_entry_p->next;
			} else {
				sesskey_engine->cache->head = del_entry_p->next;
			}
			if (del_entry_p->next != NULL) {
				del_entry_p->next->prev = del_entry_p->prev;
			} else {
				sesskey_engine->cache->tail = del_entry_p->prev;
			}
			free(del_entry_p);

		} else {
			tmp_entry_p = tmp_entry_p->next;
		}
	}
}

/*
 * Update the cache
 * Return 0 if no update
 * Return 1 if update
 */
int SESSKEY_cache_update(SESSKEY_control_p_t sesskey_engine, int tree_node_id,
		char const* bli_key, int seq) {

	SESSKEY_cache_entry_p_t tmp_entry_p;

	// first check if the pub key belongs to previous rekey interval
	if (seq < sesskey_engine->keypath->rekey_seq) {
		return 0;
	}

	// first locate the entry
	tmp_entry_p = sesskey_engine->cache->head;
	while (tmp_entry_p != NULL) {
		if (tmp_entry_p->tree_node_id == tree_node_id) {
			break;
		}
		tmp_entry_p = tmp_entry_p->next;
	}	

	// update the entry
	if (tmp_entry_p == NULL) {			// not exist
		SESSKEY_cache_insert(sesskey_engine, tree_node_id, bli_key, seq);
		return 1;
	} else {
		if (tmp_entry_p->rekey_seq < seq) {
			memcpy(tmp_entry_p->bli_key, bli_key, SESSKEY_LEN);
			tmp_entry_p->rekey_seq = seq;
			return 1;
		}
	}

	return 0;
}

/* 
 * Retrieves the blinded key from the cache based on the tree node id and seq.
 * Returns the rekey sequence number if the blinded key is found, 
 * bli_key field is returned. Returns -1 if not found.
 */
int SESSKEY_cache_retrieve(SESSKEY_control_p_t sesskey_engine, char* bli_key,
		int tree_node_id, int seq) {

	SESSKEY_cache_entry_p_t tmp_entry_p;

	// try to locate the entry
	tmp_entry_p = sesskey_engine->cache->head;	
	while (tmp_entry_p != NULL) {
		if (tmp_entry_p->tree_node_id == tree_node_id &&
				tmp_entry_p->rekey_seq == seq) {
			break;
		}
		tmp_entry_p = tmp_entry_p->next;
	}

	// returns the result if found
	if (tmp_entry_p != NULL) {
		memcpy(bli_key, tmp_entry_p->bli_key, SESSKEY_LEN);
		return tmp_entry_p->rekey_seq;
	} 

	// returns -1 if not found
	memset(bli_key, 0, SESSKEY_LEN);
	return -1;
}


/*
 * Copy all the cache elements to the keypath
 * The keypath should have all tree node ids assigned already
 * Return the largest sib_tree_node_id so as to compute the keypath
 */
void SESSKEY_cache_to_keypath(SESSKEY_control_p_t sesskey_engine) {
	SESSKEY_cache_entry_p_t tmp_entry_p;		// tmp_entry_p

	tmp_entry_p = sesskey_engine->cache->head;
	while (tmp_entry_p != NULL) {
		SESSKEY_keypath_compute(sesskey_engine, tmp_entry_p->tree_node_id, 
				tmp_entry_p->bli_key, tmp_entry_p->rekey_seq);
		
		tmp_entry_p = tmp_entry_p->next;
	}	
}

/*
 * Check if the blinded key exists in our cache
 */
int SESSKEY_cache_is_exist(SESSKEY_control_p_t sesskey_engine, 
		int tree_node_id) {

	SESSKEY_cache_entry_p_t tmp_entry_p;		// tmp_entry_p

	tmp_entry_p = sesskey_engine->cache->head;
	while (tmp_entry_p != NULL) {
		if (tmp_entry_p->tree_node_id == tree_node_id) {
			return 1;
		}	
		tmp_entry_p = tmp_entry_p->next;
	}
	return 0;
}

/*
 * Print the cache 
 */
void SESSKEY_cache_print(SESSKEY_control_p_t sesskey_engine) {
	SESSKEY_cache_entry_p_t tmp_entry_p;

	tmp_entry_p = sesskey_engine->cache->head;
	printf("\n============= CACHE START =====================\n");
	while (tmp_entry_p != NULL) {
		printf("%d ", tmp_entry_p->tree_node_id);	
		tmp_entry_p = tmp_entry_p->next;
	}
	printf("\n============= CACHE END =====================\n");
}


/*-------------------------------------------------------------------------
 *                  GROUPKEY MANAGEMENT
 *-------------------------------------------------------------------------*/
void SESSKEY_groupkey_insert(SESSKEY_control_p_t sesskey_engine, 
		char const* group_key, int seq) {
	
	SESSKEY_groupkey_entry_p_t new_entry_p, tmp_entry_p;

	// insert a new entry to the groupkey list	
	new_entry_p = (SESSKEY_groupkey_entry_p_t)
		malloc(sizeof(SESSKEY_groupkey_entry_t));
	if (new_entry_p == NULL) {
		debug_err("Unable to create keyeng groupkey entry\n");
		return;
	}

	memset(new_entry_p, 0, sizeof(SESSKEY_groupkey_entry_t));
	memcpy(new_entry_p->group_key, group_key, SESSKEY_LEN);
	new_entry_p->rekey_seq = seq;

	// insert the new entry at the list end
	new_entry_p->prev = sesskey_engine->groupkeys->tail;
	new_entry_p->next = NULL;
	if (sesskey_engine->groupkeys->head == NULL) {
		sesskey_engine->groupkeys->head = new_entry_p;
	} else {
		sesskey_engine->groupkeys->tail->next = new_entry_p;
	}
	sesskey_engine->groupkeys->tail = new_entry_p;

	sesskey_engine->groupkeys->size++;

	// remove an entry if the size exceeds a limit (say 3)
	if (sesskey_engine->groupkeys->size > 3) {
		tmp_entry_p = sesskey_engine->groupkeys->head;
		sesskey_engine->groupkeys->head = sesskey_engine->groupkeys->head->next;
		sesskey_engine->groupkeys->head->prev = NULL;
		free(tmp_entry_p);
		sesskey_engine->groupkeys->size--;
	}
}


/*
 * Retrieve the group key specified
 * Return rekey sequence number if found, -1 if not found
 */
int SESSKEY_groupkey_retrieve(SESSKEY_control_p_t sesskey_engine, 
		char* group_key, int seq) {

	SESSKEY_groupkey_entry_p_t tmp_entry_p;

	// find the group key
	if (seq == 0) {		// get the most recent one
		if (sesskey_engine->groupkeys->tail != NULL) {
			if (group_key != NULL) {	
				memcpy(group_key, sesskey_engine->groupkeys->tail->group_key, 
						SESSKEY_LEN);
			}
			return sesskey_engine->groupkeys->tail->rekey_seq;
		} 
	} else {		// get the specified one
		tmp_entry_p = sesskey_engine->groupkeys->tail;
		while (tmp_entry_p != NULL) {
			if (tmp_entry_p->rekey_seq == seq) {
				if (group_key != NULL) {	
					memcpy(group_key, tmp_entry_p->group_key, SESSKEY_LEN);
				}
				return tmp_entry_p->rekey_seq;
			}		
			tmp_entry_p = tmp_entry_p->prev;
		}	
	}

	// group key not exist
	return -1;
}


/*
 * Check if the group key list is empty.
 * 1 if empty, 0 if non-empty.
 */
int SESSKEY_groupkey_is_empty(SESSKEY_control_p_t sesskey_engine) {
	if (sesskey_engine->groupkeys->head == NULL) {
		return 1;
	}
	return 0;
}

/*
 * Check if the group key is formed given the sequence number. Returns the 
 * current sequence number (which is positive) if the group key is obtained, 
 * else return 0.
 */
int SESSKEY_groupkey_is_formed(SESSKEY_control_p_t sesskey_engine, int seq) {
	SESSKEY_groupkey_entry_p_t tmp_entry_p;
	int ret = 0;			// returned value	

	tmp_entry_p = sesskey_engine->groupkeys->tail;
	while (tmp_entry_p != NULL) {
		if (tmp_entry_p->rekey_seq == seq) {
			ret = seq;
			break;
		}
		tmp_entry_p = tmp_entry_p->prev;
	}	

	return ret;
}

/*
 * Print the groupkey list
 */
void SESSKEY_groupkey_print(SESSKEY_control_p_t sesskey_engine) {
	SESSKEY_groupkey_entry_p_t tmp_entry_p;
	int i;

	tmp_entry_p = sesskey_engine->groupkeys->head;
	printf("\n============= GROUPKEY START =====================\n");
	while (tmp_entry_p != NULL) {
		printf("seq: %d, ", tmp_entry_p->rekey_seq);
		for (i=0; i<SESSKEY_LEN; ++i) {
			printf("%d", (unsigned int)tmp_entry_p->group_key[i]);
		}
		printf("\n");
		tmp_entry_p = tmp_entry_p->next;
	}
	printf("\n============= GROUPKEY END =====================\n");
}

/*-----------------------------------------------------------------------
 *						KEY COMPUTATION
 *-----------------------------------------------------------------------*/

/*
 * Compute the secret key using Diffie-Hellman: raise the blinded key to the
 * power of secret key. Results are stored in parent_key.
 */
void SESSKEY_compute_secret(SESSKEY_control_p_t sesskey_engine, 
		char* parent_key, char const* sec_key, char const* bli_key) {

	BN_CTX *ctx;
	BIGNUM *big_sec, *big_bli, *big_p, *big_parent;

	// initialize the BIGNUM structure	
	ctx = BN_CTX_new();
	big_sec = BN_new();
	big_bli = BN_new();
	big_p = BN_new();
	big_parent = BN_new();
	
	// key computation	
	memset(parent_key, 0, SESSKEY_LEN);
	BN_bin2bn(sec_key, SESSKEY_LEN, big_sec);
	BN_bin2bn(bli_key, SESSKEY_LEN, big_bli);
	BN_bin2bn(sesskey_engine->param->p, SESSKEY_LEN, big_p);
	BN_mod_exp(big_parent, big_bli, big_sec, big_p, ctx);
	BN_bn2bin(big_parent, parent_key + SESSKEY_LEN - BN_num_bytes(big_parent));

#ifdef GAUGER       // INCREMENT THE NO_EXP COUNT
	GAUGER_inc_exp(gauger_g);
#endif

	// free the structures	
	BN_CTX_free(ctx);       
	BN_free(big_sec);
	BN_free(big_bli);
	BN_free(big_p);
	BN_free(big_parent);
}

/*
 * Compute the blinded key of a secret key.
 * Results stored in blinded_key
 */
void SESSKEY_compute_blinded(SESSKEY_control_p_t sesskey_engine, 
		char* blinded_key, char const* src_key) {
	
	BIGNUM *big_pri, *big_g, *big_p, *big_blinded;
	BN_CTX *ctx;

	// initialize the BIGNUM structures
	ctx = BN_CTX_new();
	big_pri = BN_new();
	big_g = BN_new();
	big_p = BN_new();
	big_blinded = BN_new();

	// key computation
	memset(blinded_key, 0, SESSKEY_LEN);
	BN_bin2bn(src_key, SESSKEY_LEN, big_pri);
	BN_bin2bn(sesskey_engine->param->g, SESSKEY_LEN, big_g);
	BN_bin2bn(sesskey_engine->param->p, SESSKEY_LEN, big_p);
	BN_mod_exp(big_blinded, big_g, big_pri, big_p, ctx);
	BN_bn2bin(big_blinded, blinded_key + SESSKEY_LEN - 
			BN_num_bytes(big_blinded));

#ifdef GAUGER       // INCREMENT THE NO_EXP COUNT
	GAUGER_inc_exp(gauger_g);
#endif

	// free the structures
	BN_CTX_free(ctx);
	BN_free(big_pri);
	BN_free(big_g);
	BN_free(big_p);
	BN_free(big_blinded);
}

/*-----------------------------------------------------------------------
 *                      ENCRYPTION / DECRYPTION
 *-----------------------------------------------------------------------*/

/*
 * Encrypt with the most recent group key (which is hashed before encryption)
 * Return the rekey sequence number of the group key if succeed, else -1
 * (note: len should be multiple of 8)
 */
int SESSKEY_encrypt(SESSKEY_control_p_t sesskey_engine, char const* in_data, 
		char* out_data, int len) {

	des_key_schedule ks1, ks2, ks3;
	des_cblock key1, key2, key3, iv;  // equal to unsigned char[8] 
	MD5_CTX hash_ctx;
	char hash_ret1[MD5_DIGEST_LENGTH], hash_ret2[MD5_DIGEST_LENGTH];
	char group_key[SESSKEY_LEN];
	int seq;

	if ((seq = SESSKEY_groupkey_retrieve(sesskey_engine, group_key, 0)) 
			>= 0) {

		// get the hashed value of the group key
		MD5_Init(&hash_ctx);
		MD5_Update(&hash_ctx, group_key, SESSKEY_LEN / 2);
		MD5_Final(hash_ret1, &hash_ctx);

		MD5_Init(&hash_ctx);
		MD5_Update(&hash_ctx, group_key + SESSKEY_LEN / 2, SESSKEY_LEN / 2);
		MD5_Final(hash_ret2, &hash_ctx);

		// copy the hashed value to des keys and iv
		memcpy(key1, hash_ret1, 8);
		memcpy(key2, hash_ret1 + 8, 8);
		memcpy(key3, hash_ret2, 8);
		memcpy(iv, hash_ret2 + 8, 8);

		// set the key sehedule
		des_set_key(&key1, ks1);
		des_set_key(&key2, ks2);
		des_set_key(&key3, ks3);

		// encryption
		des_ede3_cbc_encrypt(in_data, out_data, len, ks1, ks2, ks3, &iv, 
				DES_ENCRYPT);	
	} 
	return seq;
}

/*
 * Decrypt with the specified group key (which is hashed before decryption)
 * Return the rekey sequence number of the group key if succeed, else -1
 * (note: len should be multiple of 8)
 */
int SESSKEY_decrypt(SESSKEY_control_p_t sesskey_engine, char const* in_data, 
		char* out_data, int len, int seq) {

	des_key_schedule ks1, ks2, ks3;
	des_cblock key1, key2, key3, iv;  // equal to unsigned char[8] 
	char group_key[SESSKEY_LEN];
	MD5_CTX hash_ctx;
	char hash_ret1[MD5_DIGEST_LENGTH], hash_ret2[MD5_DIGEST_LENGTH];
	int ret;						// returned value

	if ((ret = SESSKEY_groupkey_retrieve(sesskey_engine, group_key, seq)) 
			>= 0) {

		// get the hashed value of the group key
		MD5_Init(&hash_ctx);
		MD5_Update(&hash_ctx, group_key, SESSKEY_LEN / 2);
		MD5_Final(hash_ret1, &hash_ctx);

		MD5_Init(&hash_ctx);
		MD5_Update(&hash_ctx, group_key + SESSKEY_LEN / 2, SESSKEY_LEN / 2);
		MD5_Final(hash_ret2, &hash_ctx);

		// copy the hashed value to des keys and iv
		memcpy(key1, hash_ret1, 8);
		memcpy(key2, hash_ret1 + 8, 8);
		memcpy(key3, hash_ret2, 8);
		memcpy(iv, hash_ret2 + 8, 8);

		// set the key sehedule
		des_set_key(&key1, ks1);
		des_set_key(&key2, ks2);
		des_set_key(&key3, ks3);

		// decryption	
		des_ede3_cbc_encrypt(in_data, out_data, len, ks1, ks2, ks3, &iv, 
				DES_DECRYPT);	
	}
	return ret;
}


/*--------------------------------------------------------------------
 *				CONFIRMATION 
 *--------------------------------------------------------------------*/

/*
 * Key confirmation.
 * Return 1 if the blinded group key received equals the one computed
 * Return 0 if the blinded key is not received / not computed
 * Return -1 if the received key and the blinded key are unequal
 */
int SESSKEY_confirm(SESSKEY_control_p_t sesskey_engine, int rekey_seq) {

	char group_key[SESSKEY_LEN];			// group key string computed
	char bkey_computed[SESSKEY_LEN];		// blinded key computed
	char bkey_received[SESSKEY_LEN];		// blinded key received

	// Double-checking:
	// 1) check if the group key is computed
	if (SESSKEY_groupkey_retrieve(sesskey_engine, group_key, rekey_seq) 
			< 0) {
		// group key is not computed
		debug("Group key is not computed.\n");
		return 0;
	}
	// 2) check if the blinded group key is received
	if (SESSKEY_cache_retrieve(sesskey_engine, bkey_received, 0, rekey_seq) 
			< 0) {
		debug("Group key is not received\n");
		return 0;
	}

	// calculate the blinded version of the group key, 
	SESSKEY_compute_blinded(sesskey_engine, bkey_computed, group_key);	

	// compare
	if (memcmp(bkey_computed, bkey_received, SESSKEY_LEN) != 0) {
/* debugging
		debug("Group key\n");
		for (i=0; i<SESSKEY_LEN; ++i) 
			printf("%d,", group_key[i]);
		debug("\nReceived bkey\n");
		for (i=0; i<SESSKEY_LEN; ++i) 
			printf("%d,", bkey_received[i]);
		printf("\n");
		debug("\nComputed bkey\n");
		for (i=0; i<SESSKEY_LEN; ++i) 
			printf("%d,", bkey_computed[i]);
		printf("\n");
debugging */
		return -1;
	} else {
		// check if zeros are obtained
		if (bkey_computed[0] == 0 && bkey_computed[1] == 0 && 
				bkey_computed[2] == 0 && bkey_computed[3] == 0) {
			debug_err("Zeros are obtained\n");
			return -1;
		}
	}

	// results are equal and valid
	return 1;
}


/*--------------------------------------------------------------------
 *				DESTROY
 *--------------------------------------------------------------------*/
/*
 * Destroy the whole route table
 */
void SESSKEY_destroy(SESSKEY_control_p_t sesskey_engine) {
	SESSKEY_keypath_entry_p_t del_keypath_entry_p;
	SESSKEY_cache_entry_p_t del_cache_entry_p;
	SESSKEY_groupkey_entry_p_t del_groupkey_entry_p;

	// free the DH param
	free(sesskey_engine->param);

	// free the keypath
	while (sesskey_engine->keypath->head != NULL) {
		del_keypath_entry_p = sesskey_engine->keypath->head;
		sesskey_engine->keypath->head = del_keypath_entry_p->next;
		free(del_keypath_entry_p);
	}
	free(sesskey_engine->keypath);
	
	// free the cache
	while (sesskey_engine->cache->head != NULL) {
		del_cache_entry_p = sesskey_engine->cache->head;
		sesskey_engine->cache->head = del_cache_entry_p->next;
		free(del_cache_entry_p);
	}
	free(sesskey_engine->cache);

	// free the groupkeys
	while (sesskey_engine->groupkeys->head  != NULL) {
		del_groupkey_entry_p = sesskey_engine->groupkeys->head;
		sesskey_engine->groupkeys->head = del_groupkey_entry_p->next;
		free(del_groupkey_entry_p);
	}
	free(sesskey_engine->groupkeys);

	// free the key engine
	free(sesskey_engine);
}
